# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines the LinearExplainer for returning explanations for linear models."""

import numpy as np

from .._internal.raw_explain.raw_explain_utils import (
    get_datamapper_and_transformed_data, transform_with_datamapper)
from ..common.aggregate import (add_explain_global_method,
                                init_aggregator_decorator)
from ..common.constants import (Attributes, Defaults, ExplainParams,
                                ExplainType, Extension, SHAPDefaults, SKLearn)
from ..common.explanation_utils import _fix_linear_explainer_shap_values
from ..common.structured_model_explainer import StructuredInitModelExplainer
from ..common.warnings_suppressor import shap_warnings_suppressor
from ..dataset.dataset_wrapper import DatasetWrapper
from ..dataset.decorator import tabular_decorator
from ..explanation.explanation import (
    _create_local_explanation, _create_raw_feats_local_explanation,
    _get_raw_explainer_create_explanation_kwargs)
from .kwargs_utils import _get_explain_global_kwargs

with shap_warnings_suppressor():
    import shap


@add_explain_global_method
class LinearExplainer(StructuredInitModelExplainer):
    available_explanations = [Extension.GLOBAL, Extension.LOCAL]
    explainer_type = Extension.GREYBOX

    """Defines the LinearExplainer for returning explanations for linear models.

    :param model: The linear model to explain as the coefficient and intercept or scikit learn model.
    :type model: (coef, intercept) or sklearn.linear_model.*
    :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
        initializing the explainer.
    :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
    :param explain_subset: List of feature indices. If specified, only selects a subset of the
        features in the evaluation dataset for explanation. The subset can be the top-k features
        from the model summary.
    :type explain_subset: list[int]
    :param features: A list of feature names.
    :type features: list[str]
    :param classes: Class names as a list of strings. The order of the class names should match
        that of the model output.  Only required if explaining classifier.
    :type classes: list[str]
    :param transformations: sklearn.compose.ColumnTransformer or a list of tuples describing the column name and
        transformer. When transformations are provided, explanations are of the features before the transformation.
        The format for a list of transformations is same as the one here:
        https://github.com/scikit-learn-contrib/sklearn-pandas.

        If you are using a transformation that is not in the list of sklearn.preprocessing transformations that
        are supported by the `interpret-community <https://github.com/interpretml/interpret-community>`_
        package, then this parameter cannot take a list of more than one column as input for the transformation.
        You can use the following sklearn.preprocessing  transformations with a list of columns since these are
        already one to many or one to one: Binarizer, KBinsDiscretizer, KernelCenterer, LabelEncoder, MaxAbsScaler,
        MinMaxScaler, Normalizer, OneHotEncoder, OrdinalEncoder, PowerTransformer, QuantileTransformer,
        RobustScaler, StandardScaler.

        Examples for transformations that work::

            [
                (["col1", "col2"], sklearn_one_hot_encoder),
                (["col3"], None) #col3 passes as is
            ]
            [
                (["col1"], my_own_transformer),
                (["col2"], my_own_transformer),
            ]

        An example of a transformation that would raise an error since it cannot be interpreted as one to many::

            [
                (["col1", "col2"], my_own_transformer)
            ]

        The last example would not work since the interpret-community package can't determine whether
        my_own_transformer gives a many to many or one to many mapping when taking a sequence of columns.
    :type transformations: sklearn.compose.ColumnTransformer or list[tuple]
    :param allow_all_transformations: Allow many to many and many to one transformations
    :type allow_all_transformations: bool
    """

    @init_aggregator_decorator
    def __init__(self, model, initialization_examples, explain_subset=None, features=None, classes=None,
                 transformations=None, allow_all_transformations=False, **kwargs):
        """Initialize the LinearExplainer.

        :param model: The linear model to explain as the coefficient and intercept or scikit learn model.
        :type model: (coef, intercept) or sklearn.linear_model.*
        :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
            initializing the explainer.
        :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param explain_subset: List of feature indices. If specified, only selects a subset of the
            features in the evaluation dataset for explanation. The subset can be the top-k features
            from the model summary.
        :type explain_subset: list[int]
        :param features: A list of feature names.
        :type features: list[str]
        :param classes: Class names as a list of strings. The order of the class names should match
            that of the model output.  Only required if explaining classifier.
        :type classes: list[str]
        :param transformations: sklearn.compose.ColumnTransformer or a list of tuples describing the column name and
            transformer. When transformations are provided, explanations are of the features before the transformation.
            The format for a list of transformations is same as the one here:
            https://github.com/scikit-learn-contrib/sklearn-pandas.

            If you are using a transformation that is not in the list of sklearn.preprocessing transformations that
            are supported by the `interpret-community <https://github.com/interpretml/interpret-community>`_
            package, then this parameter cannot take a list of more than one column as input for the transformation.
            You can use the following sklearn.preprocessing  transformations with a list of columns since these are
            already one to many or one to one: Binarizer, KBinsDiscretizer, KernelCenterer, LabelEncoder, MaxAbsScaler,
            MinMaxScaler, Normalizer, OneHotEncoder, OrdinalEncoder, PowerTransformer, QuantileTransformer,
            RobustScaler, StandardScaler.

            Examples for transformations that work::

                [
                    (["col1", "col2"], sklearn_one_hot_encoder),
                    (["col3"], None) #col3 passes as is
                ]
                [
                    (["col1"], my_own_transformer),
                    (["col2"], my_own_transformer),
                ]

            An example of a transformation that would raise an error since it cannot be interpreted as one to many::

                [
                    (["col1", "col2"], my_own_transformer)
                ]

            The last example would not work since the interpret-community package can't determine whether
            my_own_transformer gives a many to many or one to many mapping when taking a sequence of columns.
        :type transformations: sklearn.compose.ColumnTransformer or list[tuple]
        :param allow_all_transformations: Allow many to many and many to one transformations
        :type allow_all_transformations: bool
        """
        self._datamapper = None
        if transformations is not None:
            self._datamapper, initialization_examples = get_datamapper_and_transformed_data(
                examples=initialization_examples, transformations=transformations,
                allow_all_transformations=allow_all_transformations)
        super(LinearExplainer, self).__init__(model, initialization_examples, **kwargs)
        self._logger.debug('Initializing LinearExplainer')
        self._method = 'shap.linear'
        self.explainer = shap.LinearExplainer(self.model, self.initialization_examples,
                                              feature_dependence=SHAPDefaults.INDEPENDENT)
        self.explain_subset = explain_subset
        self.features = features
        self.classes = classes
        self.transformations = transformations
        self._allow_all_transformations = allow_all_transformations

    @tabular_decorator
    def explain_global(self, evaluation_examples, sampling_policy=None,
                       include_local=True, batch_size=Defaults.DEFAULT_BATCH_SIZE):
        """Explain the model globally by aggregating local explanations to global.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param sampling_policy: Optional policy for sampling the evaluation examples.  See documentation on
            SamplingPolicy for more information.
        :type sampling_policy: interpret_community.common.policy.SamplingPolicy
        :param include_local: Include the local explanations in the returned global explanation.
            If include_local is False, will stream the local explanations to aggregate to global.
        :type include_local: bool
        :param batch_size: If include_local is False, specifies the batch size for aggregating
            local explanations to global.
        :type batch_size: int
        :return: A model explanation object. It is guaranteed to be a GlobalExplanation which also has the properties
            of LocalExplanation and ExpectedValuesMixin. If the model is a classifier, it will have the properties of
            PerClassMixin.
        :rtype: DynamicGlobalExplanation
        """
        kwargs = _get_explain_global_kwargs(sampling_policy, ExplainType.SHAP_LINEAR, include_local, batch_size)
        kwargs[ExplainParams.INIT_DATA] = self.initialization_examples
        kwargs[ExplainParams.EVAL_DATA] = evaluation_examples.original_dataset_with_type
        wrapped_evals = evaluation_examples
        if self.transformations is not None:
            _, evaluation_examples = get_datamapper_and_transformed_data(
                examples=evaluation_examples,
                transformations=self.transformations,
                allow_all_transformations=self._allow_all_transformations
            )
        if isinstance(evaluation_examples, DatasetWrapper):
            kwargs[ExplainParams.NUM_FEATURES] = evaluation_examples.num_features
            evaluation_examples = evaluation_examples.original_dataset_with_type
        if len(evaluation_examples.shape) == 1:
            evaluation_examples = evaluation_examples.reshape(1, -1)
        kwargs[ExplainParams.EVAL_Y_PRED] = self.model.predict(evaluation_examples)
        if hasattr(self.model, SKLearn.PREDICT_PROBA):
            kwargs[ExplainParams.EVAL_Y_PRED_PROBA] = self.model.predict_proba(evaluation_examples)
        return self._explain_global(wrapped_evals, **kwargs)

    def _get_explain_local_kwargs(self, evaluation_examples, original_evals):
        """Get the kwargs for explain_local to create a local explanation.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param original_evals: The original data that the user passed into explain_local.
        :type original_evals: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: Args for explain_local.
        :rtype: dict
        """
        self._logger.debug('Explaining linear model')
        kwargs = {ExplainParams.METHOD: ExplainType.SHAP_LINEAR}
        if self.classes is not None:
            kwargs[ExplainParams.CLASSES] = self.classes
        kwargs[ExplainParams.FEATURES] = evaluation_examples.get_features(features=self.features)
        kwargs[ExplainParams.NUM_FEATURES] = evaluation_examples.num_features
        evaluation_examples = evaluation_examples.dataset

        shap_values = self.explainer.shap_values(evaluation_examples)
        # Temporary fix for a bug in shap for regression models
        shap_values = _fix_linear_explainer_shap_values(self.model, shap_values)
        expected_values = None
        if hasattr(self.explainer, Attributes.EXPECTED_VALUE):
            self._logger.debug('Expected values available on explainer')
            expected_values = np.array(self.explainer.expected_value)
        classification = isinstance(shap_values, list)
        # Reformat shap values result if explain_subset specified
        if self.explain_subset:
            self._logger.debug('Getting subset of shap_values')
            if classification:
                for i in range(len(shap_values)):
                    shap_values[i] = shap_values[i][:, self.explain_subset]
            else:
                shap_values = shap_values[:, self.explain_subset]
        if classification:
            kwargs[ExplainParams.MODEL_TASK] = ExplainType.CLASSIFICATION
        else:
            kwargs[ExplainParams.MODEL_TASK] = ExplainType.REGRESSION
        kwargs[ExplainParams.MODEL_TYPE] = str(type(self.model))
        kwargs[ExplainParams.LOCAL_IMPORTANCE_VALUES] = np.array(shap_values)
        kwargs[ExplainParams.EXPECTED_VALUES] = expected_values
        kwargs[ExplainParams.CLASSIFICATION] = classification
        kwargs[ExplainParams.INIT_DATA] = self.initialization_examples
        kwargs[ExplainParams.EVAL_DATA] = original_evals
        if len(evaluation_examples.shape) == 1:
            evaluation_examples = evaluation_examples.reshape(1, -1)
        kwargs[ExplainParams.EVAL_Y_PRED] = self.model.predict(evaluation_examples)
        if hasattr(self.model, SKLearn.PREDICT_PROBA):
            kwargs[ExplainParams.EVAL_Y_PRED_PROBA] = self.model.predict_proba(evaluation_examples)
        return kwargs

    @tabular_decorator
    def explain_local(self, evaluation_examples):
        """Explain the model by using SHAP's linear explainer.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: DatasetWrapper
        :return: A model explanation object. It is guaranteed to be a LocalExplanation which also has the properties
            of ExpectedValuesMixin. If the model is a classifier, it will have the properties of the ClassesMixin.
        :rtype: DynamicLocalExplanation
        """
        original_evals = evaluation_examples.original_dataset_with_type
        if self._datamapper is not None:
            evaluation_examples = transform_with_datamapper(evaluation_examples, self._datamapper)

        kwargs = self._get_explain_local_kwargs(evaluation_examples, original_evals)
        explanation = _create_local_explanation(**kwargs)

        if self._datamapper is None:
            return explanation
        else:
            # if transformations have been passed, then return raw features explanation
            raw_kwargs = _get_raw_explainer_create_explanation_kwargs(kwargs=kwargs)
            return _create_raw_feats_local_explanation(explanation, feature_maps=[self._datamapper.feature_map],
                                                       features=self.features, **raw_kwargs)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines an explainable lightgbm model."""

import inspect
import json
import logging

from packaging import version
from scipy.sparse import issparse

from ...common.constants import (ExplainableModelType, Extension,
                                 LightGBMSerializationConstants,
                                 ShapValuesOutput)
from ...common.warnings_suppressor import shap_warnings_suppressor
from .explainable_model import (BaseExplainableModel, _clean_doc,
                                _get_initializer_args)
from .tree_model_utils import (_expected_values_tree_surrogate,
                               _explain_local_tree_surrogate)

with shap_warnings_suppressor():
    import shap


try:
    import lightgbm
    from lightgbm import Booster, LGBMClassifier, LGBMRegressor
    if (version.parse(lightgbm.__version__) <= version.parse('2.2.1')):
        print("Using older than supported version of lightgbm, please upgrade to version greater than 2.2.1")
except ImportError:
    print("Could not import lightgbm, required if using LGBMExplainableModel")


DEFAULT_RANDOM_STATE = 123
_N_FEATURES = '_n_features'
_N_CLASSES = '_n_classes'
NUM_ITERATIONS = 'num_iterations'
_FITTED = 'fitted_'


class _LGBMFunctionWrapper(object):
    """Decorate the predict method, temporary workaround for sparse case until TreeExplainer support is added.

    :param function: The prediction function to wrap.
    :type function: function
    """

    def __init__(self, function):
        """Wraps a function to reshape the input data.

        :param function: The prediction function to wrap.
        :type function: function
        """
        self._function = function

    def predict_wrapper(self, X, *args, **kwargs):
        """Wraps a prediction function from lightgbm learner.

        If version is ==3.0.0, densifies the input dataset.

        :param X: The model evaluation examples.
        :type X: numpy.array
        :return: Prediction result.
        :rtype: numpy.array
        """
        if issparse(X):
            X = X.toarray()
        return self._function(X, *args, **kwargs)


class _SparseTreeExplainer(object):

    """Wraps the lightgbm model to enable sparse feature contributions.

    If version is >=3.1.0, runs on sparse input data by calling predict function directly.

    :param lgbm: The lightgbm model to wrap.
    :type lgbm: LGBMModel
    :param tree_explainer: The tree_explainer used for dense data.
    :type tree_explainer: shap.TreeExplainer
    """

    def __init__(self, lgbm, tree_explainer):
        """Wraps the lightgbm model to enable sparse feature contributions.

        :param lgbm: The lightgbm model to wrap.
        :type lgbm: LGBMModel
        :param tree_explainer: The tree_explainer used for dense data.
        :type tree_explainer: shap.TreeExplainer
        """
        self._lgbm = lgbm
        self._tree_explainer = tree_explainer
        self._num_iters = -1
        # Get the number of iterations trained for from the booster
        if hasattr(self._lgbm._Booster, 'params'):
            if NUM_ITERATIONS in self._lgbm._Booster.params:
                self._num_iters = self._lgbm._Booster.params[NUM_ITERATIONS]
        # If best iteration specified, use that
        if self._lgbm._best_iteration is not None:
            self._num_iters = self._lgbm._best_iteration
        self.expected_value = None

    def shap_values(self, X):
        """Calls lightgbm predict directly for sparse case.

        If lightgbm version is >=3.1.0, runs on sparse input data
        by calling predict function directly with pred_contrib=True.
        Uses tree explainer for dense input data.

        :param X: The model evaluation examples.
        :type X: numpy.array or scipy.sparse.csr_matrix
        :return: The feature importance values.
        :rtype: numpy.array, scipy.sparse or list of scipy.sparse
        """
        if issparse(X):
            shap_values = self._lgbm.predict(X,
                                             num_iteration=self._num_iters,
                                             pred_contrib=True)
            if isinstance(shap_values, list):
                shape = shap_values[0].shape
                self.expected_value = shap_values[0][0, shape[1] - 1]
                for idx, class_values in enumerate(shap_values):
                    shap_values[idx] = class_values[:, :shape[1] - 1]
            else:
                shape = shap_values.shape
                self.expected_value = shap_values[0, shape[1] - 1]
                shap_values = shap_values[:, :shape[1] - 1]
        else:
            shap_values = self._tree_explainer.shap_values(X)
            self.expected_value = self._tree_explainer.expected_value
        return shap_values


class LGBMExplainableModel(BaseExplainableModel):
    available_explanations = [Extension.GLOBAL, Extension.LOCAL]
    explainer_type = Extension.GLASSBOX

    """LightGBM (fast, high performance framework based on decision tree) explainable model.

    Please see documentation for more details: https://github.com/Microsoft/LightGBM

    Additional arguments to LightGBMClassifier and LightGBMRegressor can be passed through kwargs.

    :param multiclass: Set to true to generate a multiclass model.
    :type multiclass: bool
    :param random_state: Int to seed the model.
    :type random_state: int
    :param shap_values_output: The type of the output from explain_local when using TreeExplainer.
        Currently only types 'default', 'probability' and 'teacher_probability' are supported.  If
        'probability' is specified, then we approximately scale the raw log-odds values from the
        TreeExplainer to probabilities.
    :type shap_values_output: interpret_community.common.constants.ShapValuesOutput
    :param classification: Indicates if this is a classification or regression explanation.
    :type classification: bool
    """

    def __init__(self, multiclass=False, random_state=DEFAULT_RANDOM_STATE,
                 shap_values_output=ShapValuesOutput.DEFAULT, classification=True, **kwargs):
        """Initialize the LightGBM Model.

        Additional arguments to LightGBMClassifier and LightGBMRegressor can be passed through kwargs.

        :param multiclass: Set to true to generate a multiclass model.
        :type multiclass: bool
        :param random_state: Int to seed the model.
        :type random_state: int
        :param shap_values_output: The type of the output from explain_local when using TreeExplainer.
            Currently only types 'default', 'probability' and 'teacher_probability' are supported.  If
            'probability' is specified, then we approximately scale the raw log-odds values from the
            TreeExplainer to probabilities.
        :type shap_values_output: interpret_community.common.constants.ShapValuesOutput
        :param classification: Indicates if this is a classification or regression explanation.
        :type classification: bool
        """
        self.multiclass = multiclass
        initializer_args = _get_initializer_args(kwargs)
        if self.multiclass:
            initializer = LGBMClassifier
        else:
            initializer = LGBMRegressor
        self._lgbm = initializer(random_state=random_state, **initializer_args)
        super(LGBMExplainableModel, self).__init__(**kwargs)
        self._logger.debug('Initializing LGBMExplainableModel')
        self._method = 'lightgbm'
        self._tree_explainer = None
        self._shap_values_output = shap_values_output
        self._classification = classification

    try:
        __init__.__doc__ = (__init__.__doc__ +
                            '\nIf multiclass=True, uses the parameters for LGBMClassifier:\n' +
                            _clean_doc(LGBMClassifier.__init__.__doc__) +
                            '\nOtherwise, if multiclass=False, uses the parameters for LGBMRegressor:\n' +
                            _clean_doc(LGBMRegressor.__init__.__doc__))
    except Exception:
        pass

    def fit(self, dataset, labels, **kwargs):
        """Call lightgbm fit to fit the explainable model.

        :param dataset: The dataset to train the model on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param labels: The labels to train the model on.
        :type labels: numpy.array
        """
        self._lgbm.fit(dataset, labels, **kwargs)

    try:
        fit.__doc__ = (fit.__doc__ +
                       '\nIf multiclass=True, uses the parameters for LGBMClassifier:\n' +
                       _clean_doc(LGBMClassifier.fit.__doc__) +
                       '\nOtherwise, if multiclass=False, uses the parameters for LGBMRegressor:\n' +
                       _clean_doc(LGBMRegressor.fit.__doc__))
    except Exception:
        pass

    def predict(self, dataset, **kwargs):
        """Call lightgbm predict to predict labels using the explainable model.

        :param dataset: The dataset to predict on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The predictions of the model.
        :rtype: list
        """
        return self._lgbm.predict(dataset, **kwargs)

    try:
        predict.__doc__ = (predict.__doc__ +
                           '\nIf multiclass=True, uses the parameters for LGBMClassifier:\n' +
                           _clean_doc(LGBMClassifier.predict.__doc__) +
                           '\nOtherwise, if multiclass=False, uses the parameters for LGBMRegressor:\n' +
                           _clean_doc(LGBMRegressor.predict.__doc__))
    except Exception:
        pass

    def predict_proba(self, dataset, **kwargs):
        """Call lightgbm predict_proba to predict probabilities using the explainable model.

        :param dataset: The dataset to predict probabilities on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The predictions of the model.
        :rtype: list
        """
        if self.multiclass:
            return self._lgbm.predict_proba(dataset, **kwargs)
        else:
            raise Exception("predict_proba not supported for regression or binary classification dataset")

    try:
        predict_proba.__doc__ = (predict_proba.__doc__ +
                                 '\nIf multiclass=True, uses the parameters for LGBMClassifier:\n' +
                                 _clean_doc(LGBMClassifier.predict_proba.__doc__) +
                                 '\nOtherwise predict_proba is not supported for ' +
                                 'regression or binary classification.\n')
    except Exception:
        pass

    def explain_global(self, **kwargs):
        """Call lightgbm feature importances to get the global feature importances from the explainable model.

        :return: The global explanation of feature importances.
        :rtype: numpy.ndarray
        """
        return self._lgbm.feature_importances_

    def _init_tree_explainer(self):
        """Creates the TreeExplainer.

        Includes a temporary fix for lightgbm 3.0 by wrapping predict method
        for sparse case to output dense data.
        Includes another temporary fix for lightgbm >= 3.1 to call predict
        function directly for sparse input data until shap TreeExplainer
        support is added.
        """
        if self._tree_explainer is None:
            self._tree_explainer = shap.TreeExplainer(self._lgbm)
            if version.parse('3.1.0') <= version.parse(lightgbm.__version__):
                self._tree_explainer = _SparseTreeExplainer(self._lgbm, self._tree_explainer)
            elif version.parse('3.0.0') == version.parse(lightgbm.__version__):
                wrapper = _LGBMFunctionWrapper(self._tree_explainer.model.original_model.predict)
                self._tree_explainer.model.original_model.predict = wrapper.predict_wrapper

    def explain_local(self, evaluation_examples, probabilities=None, **kwargs):
        """Use TreeExplainer to get the local feature importances from the trained explainable model.

        :param evaluation_examples: The evaluation examples to compute local feature importances for.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param probabilities: If output_type is probability, can specify the teacher model's
            probability for scaling the shap values.
        :type probabilities: numpy.ndarray
        :return: The local explanation of feature importances.
        :rtype: Union[list, numpy.ndarray]
        """
        self._init_tree_explainer()
        return _explain_local_tree_surrogate(self._lgbm, evaluation_examples, self._tree_explainer,
                                             self._shap_values_output, self._classification,
                                             probabilities, self.multiclass)

    @property
    def expected_values(self):
        """Use TreeExplainer to get the expected values.

        :return: The expected values of the LightGBM tree model.
        :rtype: list
        """
        self._init_tree_explainer()
        return _expected_values_tree_surrogate(self._lgbm, self._tree_explainer, self._shap_values_output,
                                               self._classification, self.multiclass)

    @property
    def model(self):
        """Retrieve the underlying model.

        :return: The lightgbm model, either classifier or regressor.
        :rtype: Union[LGBMClassifier, LGBMRegressor]
        """
        return self._lgbm

    @staticmethod
    def explainable_model_type():
        """Retrieve the model type.

        :return: Tree explainable model type.
        :rtype: ExplainableModelType
        """
        return ExplainableModelType.TREE_EXPLAINABLE_MODEL_TYPE

    def _save(self):
        """Return a string dictionary representation of the LGBMExplainableModel.

        :return: A serialized dictionary representation of the LGBMExplainableModel.
        :rtype: dict
        """
        properties = {}
        # Save all of the properties
        for key, value in self.__dict__.items():
            if key in LightGBMSerializationConstants.nonify_properties:
                properties[key] = None
            elif key in LightGBMSerializationConstants.save_properties:
                # Save booster model to string representation
                # This is not recommended but can be necessary to get around pickle being not secure
                # See here for more info:
                # https://github.com/Microsoft/LightGBM/issues/1942
                # https://github.com/Microsoft/LightGBM/issues/1217
                properties[key] = value.booster_.model_to_string()
            else:
                properties[key] = json.dumps(value)
        # Need to add _n_features
        properties[_N_FEATURES] = self._lgbm._n_features
        # And if classification case need to add _n_classes
        if self.multiclass:
            properties[_N_CLASSES] = self._lgbm._n_classes
        if hasattr(self._lgbm, _FITTED):
            properties[_FITTED] = json.dumps(getattr(self._lgbm, _FITTED))
        return properties

    @staticmethod
    def _load(properties):
        """Load a LGBMExplainableModel from the given properties.

        :param properties: A serialized dictionary representation of the LGBMExplainableModel.
        :type properties: dict
        :return: The deserialized LGBMExplainableModel.
        :rtype: interpret_community.mimic.models.LGBMExplainableModel
        """
        # create the LGBMExplainableModel without any properties using the __new__ function, similar to pickle
        lgbm_model = LGBMExplainableModel.__new__(LGBMExplainableModel)
        # Get _n_features
        _n_features = properties.pop(_N_FEATURES)
        # If classification case get _n_classes
        if json.loads(properties[LightGBMSerializationConstants.MULTICLASS]):
            _n_classes = properties.pop(_N_CLASSES)
        fitted_ = None
        if _FITTED in properties:
            fitted_ = json.loads(properties[_FITTED])
        elif version.parse('3.3.1') <= version.parse(lightgbm.__version__):
            # If deserializing older model in newer version set this to true to prevent errors on calls
            fitted_ = True
        # load all of the properties
        for key, value in properties.items():
            # Regenerate the properties on the fly
            if key in LightGBMSerializationConstants.nonify_properties:
                if key == LightGBMSerializationConstants.LOGGER:
                    parent = logging.getLogger(__name__)
                    lightgbm_identity = json.loads(properties[LightGBMSerializationConstants.IDENTITY])
                    lgbm_model.__dict__[key] = parent.getChild(lightgbm_identity)
                elif key == LightGBMSerializationConstants.TREE_EXPLAINER:
                    lgbm_model.__dict__[key] = None
                else:
                    raise Exception("Unknown nonify key on deserialize in LightGBMExplainableModel: {}".format(key))
            elif key in LightGBMSerializationConstants.save_properties:
                # Load the booster from file and re-create the LGBMClassifier or LGBMRegressor
                # This is not recommended but can be necessary to get around pickle being not secure
                # See here for more info:
                # https://github.com/Microsoft/LightGBM/issues/1942
                # https://github.com/Microsoft/LightGBM/issues/1217
                booster_args = {LightGBMSerializationConstants.MODEL_STR: value}
                is_multiclass = json.loads(properties[LightGBMSerializationConstants.MULTICLASS])
                if is_multiclass:
                    objective = LightGBMSerializationConstants.MULTICLASS
                else:
                    objective = LightGBMSerializationConstants.REGRESSION
                if LightGBMSerializationConstants.MODEL_STR in inspect.getargspec(Booster).args:
                    extras = {LightGBMSerializationConstants.OBJECTIVE: objective}
                    lgbm_booster = Booster(**booster_args, params=extras)
                else:
                    # For backwards compatibility with older versions of lightgbm
                    booster_args[LightGBMSerializationConstants.OBJECTIVE] = objective
                    lgbm_booster = Booster(params=booster_args)
                if is_multiclass:
                    new_lgbm = LGBMClassifier()
                    new_lgbm._Booster = lgbm_booster
                    new_lgbm._n_classes = _n_classes
                else:
                    new_lgbm = LGBMRegressor()
                    new_lgbm._Booster = lgbm_booster
                # Specify fitted_ for newer versions of lightgbm on deserialize
                if fitted_ is not None:
                    new_lgbm.fitted_ = fitted_
                new_lgbm._n_features = _n_features
                lgbm_model.__dict__[key] = new_lgbm
            elif key in LightGBMSerializationConstants.enum_properties:
                # NOTE: If more enums added in future, will need to handle this differently
                lgbm_model.__dict__[key] = ShapValuesOutput(json.loads(value))
            else:
                lgbm_model.__dict__[key] = json.loads(value)
        return lgbm_model



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Reimports a helpful dataset wrapper to allow operations such as summarizing data, taking the subset or sampling."""

from ml_wrappers.dataset import CustomTimestampFeaturizer, DatasetWrapper

__all__ = ['CustomTimestampFeaturizer', 'DatasetWrapper']



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines explanation policies."""

from .chained_identity import ChainedIdentity
from .constants import Defaults


class SamplingPolicy(ChainedIdentity):
    """Defines the sampling policy for downsampling the evaluation examples.

    The policy is a set of parameters that can be tuned to speed up or improve the accuracy of the
    explain_model function during sampling.

    :param allow_eval_sampling: Default to 'False'. Specify whether to allow sampling of evaluation data.
        If 'True', cluster the evaluation data and determine the optimal number
        of points for sampling. Set to 'True' to speed up the process when the
        evaluation data set is large and you only want to generate model
        summary info.
    :type allow_eval_sampling: bool
    :param max_dim_clustering: Default to 50 and only take effect when 'allow_eval_sampling' is
        set to 'True'. Specify the dimensionality to reduce the evaluation data before clustering
        for sampling. When doing sampling to determine how aggressively to downsample without getting poor
        explanation results uses a heuristic to find the optimal number of clusters. Since
        KMeans performs poorly on high dimensional data PCA or Truncated SVD is first run to
        reduce the dimensionality, which is followed by finding the optimal k by running
        KMeans until a local minimum is reached as determined by computing the silhouette
        score, reducing k each time.
    :type max_dim_clustering: int
    :param sampling_method: The sampling method for determining how much to downsample the evaluation data by.
        If allow_eval_sampling is True, the evaluation data is downsampled to a max_threshold, and then this
        heuristic is used to determine how much more to downsample the evaluation data without losing accuracy
        on the calculated feature importance values.  By default, this is set to hdbscan, but you can
        also specify kmeans.  With hdbscan the number of clusters is automatically determined and multiplied by
        a threshold.  With kmeans, the optimal number of clusters is found by running KMeans until the maximum
        silhouette score is calculated, with k halved each time.
    :type sampling_method: str
    :rtype: dict
    :return: The arguments for the sampling policy
    """

    def __init__(self, allow_eval_sampling=False, max_dim_clustering=Defaults.MAX_DIM,
                 sampling_method=Defaults.HDBSCAN, **kwargs):
        """Initialize the SamplingPolicy.

        :param allow_eval_sampling: Default to 'False'. Specify whether to allow sampling of evaluation data.
            If 'True', cluster the evaluation data and determine the optimal number
            of points for sampling. Set to 'True' to speed up the process when the
            evaluation data set is large and you only want to generate model
            summary info.
        :type allow_eval_sampling: bool
        :param max_dim_clustering: Default to 50 and only take effect when 'allow_eval_sampling' is
            set to 'True'. Specify the dimensionality to reduce the evaluation data before clustering
            for sampling. When doing sampling to determine how aggressively to downsample without getting poor
            explanation results uses a heuristic to find the optimal number of clusters. Since
            KMeans performs poorly on high dimensional data PCA or Truncated SVD is first run to
            reduce the dimensionality, which is followed by finding the optimal k by running
            KMeans until a local minimum is reached as determined by computing the silhouette
            score, reducing k each time.
        :type max_dim_clustering: int
        :param sampling_method: The sampling method for determining how much to downsample the evaluation data by.
            If allow_eval_sampling is True, the evaluation data is downsampled to a max_threshold, and then this
            heuristic is used to determine how much more to downsample the evaluation data without losing accuracy
            on the calculated feature importance values.  By default, this is set to hdbscan, but the user can
            also specify kmeans. With hdbscan the number of clusters is automatically determined and multiplied by
            a threshold. With kmeans, the optimal number of clusters is found by running KMeans until the maximum
            silhouette score is calculated, with k halved each time. For more information about hbdscan, see
            https://github.com/scikit-learn-contrib/hdbscan.
        :type sampling_method: str
        :rtype: dict
        :return: The arguments for the sampling policy
        """
        super(SamplingPolicy, self).__init__(**kwargs)
        self._allow_eval_sampling = allow_eval_sampling
        self._max_dim_clustering = max_dim_clustering
        self._sampling_method = sampling_method

    @property
    def allow_eval_sampling(self):
        """Get whether to allow sampling of evaluation data.

        :return: Whether to allow sampling of evaluation data.
        :rtype: bool
        """
        return self._allow_eval_sampling

    @property
    def max_dim_clustering(self):
        """Get the dimensionality to reduce the evaluation data before clustering for sampling.

        :return: The dimensionality to reduce the evaluation data before clustering for sampling.
        :rtype: int
        """
        return self._max_dim_clustering

    @property
    def sampling_method(self):
        """Get the sampling method for determining how much to downsample the evaluation data by.

        :return: The sampling method for determining how much to downsample the evaluation data by.
        :rtype: str
        """
        return self._sampling_method



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines the base API for explainable models."""
import logging
from abc import ABC, abstractmethod

from ...common.chained_identity import ChainedIdentity


def _get_initializer_args(surrogate_init_args):
    """Return a list of args to default values for the given function that are in the given argument dict.

    :param function: The function to retrieve the arguments from.
    :type function: Function
    :param surrogate_init_args: The arguments to initialize the surrogate model.
    :type surrogate_init_args: dict
    :return: A mapping from argument name to value for the surrogate model.
    :rtype: dict
    """
    # List of known args that child and base explainer are known to support
    base_model_args = ['_ident', '_parent_logger']
    # Pass all other args to underlying model
    return dict([(arg, surrogate_init_args.pop(arg))
                 for arg in surrogate_init_args.copy()
                 if arg not in base_model_args])


def _clean_doc(doc):
    return doc.replace('-', '').replace(':term:', '').replace('**', '')


class BaseExplainableModel(ABC, ChainedIdentity):
    """The base class for models that can be explained."""

    def __init__(self, **kwargs):
        """Initialize the Explainable Model."""
        super(BaseExplainableModel, self).__init__(**kwargs)
        self._logger.debug('Initializing BaseExplainableModel')

    @abstractmethod
    def fit(self, **kwargs):
        """Abstract method to fit the explainable model."""
        pass

    @abstractmethod
    def predict(self, dataset, **kwargs):
        """Abstract method to predict labels using the explainable model."""
        pass

    @abstractmethod
    def predict_proba(self, dataset, **kwargs):
        """Abstract method to predict probabilities using the explainable model."""
        pass

    @abstractmethod
    def explain_global(self, **kwargs):
        """Abstract method to get the global feature importances from the trained explainable model."""
        pass

    @abstractmethod
    def explain_local(self, evaluation_examples, **kwargs):
        """Abstract method to get the local feature importances from the trained explainable model."""
        pass

    @property
    @abstractmethod
    def expected_values(self):
        """Abstract property to get the expected values."""
        pass

    @property
    @abstractmethod
    def model(self):
        """Abstract property to get the underlying model."""
        pass

    @staticmethod
    def explainable_model_type():
        """Retrieve the model type."""
        pass

    def __getstate__(self):
        """Influence how SGDExplainableModel is pickled.

        Removes logger which is not serializable.

        :return state: The state to be pickled, with logger removed.
        :rtype: dict
        """
        odict = self.__dict__.copy()
        del odict['_logger']
        return odict

    def __setstate__(self, state):
        """Influence how SGDExplainableModel is unpickled.

        Re-adds logger which is not serializable.

        :param dict: A dictionary of deserialized state.
        :type dict: dict
        """
        self.__dict__.update(state)
        self._logger = logging.getLogger(__name__)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines an explainable tree model."""

from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from ...common.constants import (ExplainableModelType, Extension,
                                 ShapValuesOutput)
from ...common.explanation_utils import _get_dense_examples
from ...common.warnings_suppressor import shap_warnings_suppressor
from .explainable_model import (BaseExplainableModel, _clean_doc,
                                _get_initializer_args)
from .tree_model_utils import (_expected_values_tree_surrogate,
                               _explain_local_tree_surrogate)

with shap_warnings_suppressor():
    import shap

DEFAULT_RANDOM_STATE = 123


class DecisionTreeExplainableModel(BaseExplainableModel):
    available_explanations = [Extension.GLOBAL, Extension.LOCAL]
    explainer_type = Extension.GLASSBOX

    """Decision Tree explainable model.

    :param multiclass: Set to true to generate a multiclass model.
    :type multiclass: bool
    :param random_state: Int to seed the model.
    :type random_state: int
    :param shap_values_output: The type of the output from explain_local when using TreeExplainer.
        Currently only types 'default', 'probability' and 'teacher_probability' are supported.  If
        'probability' is specified, then we approximately scale the raw log-odds values from the
        TreeExplainer to probabilities.
    :type shap_values_output: interpret_community.common.constants.ShapValuesOutput
    :param classification: Indicates if this is a classification or regression explanation.
    :type classification: bool
    """

    def __init__(self, multiclass=False, random_state=DEFAULT_RANDOM_STATE,
                 shap_values_output=ShapValuesOutput.DEFAULT, classification=True, **kwargs):
        """Initialize the DecisionTreeExplainableModel.

        :param multiclass: Set to true to generate a multiclass model.
        :type multiclass: bool
        :param random_state: Int to seed the model.
        :type random_state: int
        :param shap_values_output: The type of the output from explain_local when using TreeExplainer.
            Currently only types 'default', 'probability' and 'teacher_probability' are supported.  If
            'probability' is specified, then we approximately scale the raw log-odds values from the
            TreeExplainer to probabilities.
        :type shap_values_output: interpret_community.common.constants.ShapValuesOutput
        :param classification: Indicates if this is a classification or regression explanation.
        :type classification: bool
        """
        self.multiclass = multiclass
        self.random_state = random_state
        kwargs['random_state'] = random_state
        if self.multiclass:
            initializer = DecisionTreeClassifier
        else:
            initializer = DecisionTreeRegressor
        initializer_args = _get_initializer_args(kwargs)
        self._tree = initializer(**initializer_args)
        super(DecisionTreeExplainableModel, self).__init__(**kwargs)
        self._logger.debug('Initializing DecisionTreeExplainableModel')
        self._method = 'tree'
        self._tree_explainer = None
        self._shap_values_output = shap_values_output
        self._classification = classification

    __init__.__doc__ = (__init__.__doc__ +
                        '\nIf multiclass=True, uses the parameters for DecisionTreeClassifier:\n' +
                        _clean_doc(DecisionTreeClassifier.__doc__) +
                        '\nOtherwise, if multiclass=False, uses the parameters for DecisionTreeRegressor:\n' +
                        _clean_doc(DecisionTreeRegressor.__doc__))

    def fit(self, dataset, labels, **kwargs):
        """Call tree fit to fit the explainable model.

        :param dataset: The dataset to train the model on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param labels: The labels to train the model on.
        :type labels: numpy.array
        """
        self._tree.fit(dataset, labels, **kwargs)

    fit.__doc__ = (fit.__doc__ +
                   '\nIf multiclass=True, uses the parameters for DecisionTreeClassifier:\n' +
                   _clean_doc(DecisionTreeClassifier.fit.__doc__) +
                   '\nOtherwise, if multiclass=False, uses the parameters for DecisionTreeRegressor:\n' +
                   _clean_doc(DecisionTreeRegressor.fit.__doc__))

    def predict(self, dataset, **kwargs):
        """Call tree predict to predict labels using the explainable model.

        :param dataset: The dataset to predict on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The predictions of the model.
        :rtype: list
        """
        return self._tree.predict(dataset)

    predict.__doc__ = (predict.__doc__ +
                       '\nIf multiclass=True, uses the parameters for DecisionTreeClassifier:\n' +
                       _clean_doc(DecisionTreeClassifier.predict.__doc__) +
                       '\nOtherwise, if multiclass=False, uses the parameters for DecisionTreeRegressor:\n' +
                       _clean_doc(DecisionTreeRegressor.predict.__doc__))

    def predict_proba(self, dataset, **kwargs):
        """Call tree predict_proba to predict probabilities using the explainable model.

        :param dataset: The dataset to predict probabilities on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The predictions of the model.
        :rtype: list
        """
        if self.multiclass:
            return self._tree.predict_proba(dataset)
        else:
            raise Exception('predict_proba not supported for regression or binary classification dataset')

    predict_proba.__doc__ = (predict_proba.__doc__ +
                             '\nIf multiclass=True, uses the parameters for DecisionTreeClassifier:\n' +
                             _clean_doc(DecisionTreeClassifier.predict_proba.__doc__) +
                             '\nOtherwise predict_proba is not supported for regression or binary classification.\n')

    def explain_global(self, **kwargs):
        """Call tree model feature importances to get the global feature importances from the tree surrogate model.

        :return: The global explanation of feature importances.
        :rtype: list
        """
        return self._tree.feature_importances_

    def explain_local(self, evaluation_examples, probabilities=None, **kwargs):
        """Use TreeExplainer to get the local feature importances from the trained explainable model.

        :param evaluation_examples: The evaluation examples to compute local feature importances for.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param probabilities: If output_type is probability, can specify the teacher model's
            probability for scaling the shap values.
        :type probabilities: numpy.ndarray
        :return: The local explanation of feature importances.
        :rtype: Union[list, numpy.ndarray]
        """
        if self._tree_explainer is None:
            self._tree_explainer = shap.TreeExplainer(self._tree)
        evaluation_examples = _get_dense_examples(evaluation_examples)
        return _explain_local_tree_surrogate(self._tree, evaluation_examples, self._tree_explainer,
                                             self._shap_values_output, self._classification,
                                             probabilities, self.multiclass)

    @property
    def expected_values(self):
        """Use TreeExplainer to get the expected values.

        :return: The expected values of the decision tree tree model.
        :rtype: list
        """
        if self._tree_explainer is None:
            self._tree_explainer = shap.TreeExplainer(self._tree)
        return _expected_values_tree_surrogate(self._tree, self._tree_explainer, self._shap_values_output,
                                               self._classification, self.multiclass)

    @property
    def model(self):
        """Retrieve the underlying model.

        :return: The decision tree model, either classifier or regressor.
        :rtype: Union[sklearn.tree.DecisionTreeClassifier, sklearn.tree.DecisionTreeRegressor]
        """
        return self._tree

    @staticmethod
    def explainable_model_type():
        """Retrieve the model type.

        :return: Tree explainable model type.
        :rtype: interpret_community.common.constants.ExplainableModelType
        """
        return ExplainableModelType.TREE_EXPLAINABLE_MODEL_TYPE



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines utilities for tree-based explainable models."""

from scipy.special import expit

from ...common.constants import ShapValuesOutput
from ...common.explanation_utils import _scale_tree_shap


def _explain_local_tree_surrogate(tree_model, evaluation_examples, tree_explainer,
                                  shap_values_output, classification, probabilities,
                                  multiclass):
    """Locally explains the tree-based surrogate model.

    :param tree_model: A tree-based model.
    :type tree_model: Tree-based model with scikit-learn predict and predict_proba API.
    :param evaluation_examples: The evaluation examples to compute local feature importances for.
    :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
    :param tree_explainer: Tree explainer for the tree-based model.
    :type tree_explainer: TreeExplainer
    :param shap_values_output: The type of the output from explain_local when using TreeExplainer.
        Currently only types 'default', 'probability' and 'teacher_probability' are supported.  If
        'probability' is specified, then we approximately scale the raw log-odds values from the
        TreeExplainer to probabilities.
    :type shap_values_output: interpret_community.common.constants.ShapValuesOutput
    :param classification: Indicates if this is a classification or regression explanation.
    :type classification: bool
    :param probabilities: If output_type is probability, can specify the teacher model's
        probability for scaling the shap values.
        Note for regression case this will just be the output values of the teacher model.
    :type probabilities: numpy.ndarray
    :param multiclass: True if the tree_model is a multiclass model.
    :type multiclass: bool
    :return: The local explanation of feature importances.
    :rtype: Union[list, numpy.ndarray]
    """
    if len(evaluation_examples.shape) == 1:
        evaluation_examples = evaluation_examples.reshape(1, -1)
    # Note: For binary and multiclass case the expected values and shap values are in the units
    # of the raw predictions from the underlying model.
    # In binary case, we are using regressor on logit.  In multiclass case, shap TreeExplainer
    # outputs the margin instead of probabilities.
    shap_values = tree_explainer.shap_values(evaluation_examples)
    is_probability = shap_values_output == ShapValuesOutput.PROBABILITY
    is_teacher_probability = shap_values_output == ShapValuesOutput.TEACHER_PROBABILITY
    if is_probability or is_teacher_probability:
        expected_values = tree_explainer.expected_value
        if classification:
            expected_values = expit(expected_values)
            if probabilities is None:
                if is_teacher_probability:
                    raise Exception("Probabilities not specified for output type 'teacher_probability'")
                if multiclass:
                    probabilities = tree_model.predict_proba(evaluation_examples)
                else:
                    probabilities = expit(tree_model.predict(evaluation_examples))
                    probabilities = probabilities.reshape((probabilities.shape[0], 1))
            shap_values = _scale_tree_shap(shap_values, expected_values, probabilities)
        elif is_teacher_probability:
            # In regression case, the values are in terms of the surrogate model, so we transform to teacher model
            # note this is not in terms of probabilities, might need to figure out a better param name
            shap_values = _scale_tree_shap(shap_values, expected_values, probabilities)
    return shap_values


def _expected_values_tree_surrogate(tree_model, tree_explainer, shap_values_output, classification, multiclass):
    """Use TreeExplainer to get the expected values.

    :param tree_model: A tree-based model.
    :type tree_model: Tree-based model with scikit-learn predict and predict_proba API.
    :param tree_explainer: Tree explainer for the tree-based model.
    :type tree_explainer: TreeExplainer
    :param shap_values_output: The type of the output from explain_local when using TreeExplainer.
        Currently only types 'default', 'probability' and 'teacher_probability' are supported.  If
        'probability' is specified, then we approximately scale the raw log-odds values from the
        TreeExplainer to probabilities.
    :type shap_values_output: interpret_community.common.constants.ShapValuesOutput
    :param classification: Indicates if this is a classification or regression explanation.
    :type classification: bool
    :param multiclass: True if the tree_model is a multiclass model.
    :type multiclass: bool
    :return: The expected values of the tree-based model.
    :rtype: list
    """
    expected_values = tree_explainer.expected_value
    if classification:
        if not multiclass:
            expected_values = [-expected_values, expected_values]
        is_probability = shap_values_output == ShapValuesOutput.PROBABILITY
        is_teacher_probability = shap_values_output == ShapValuesOutput.TEACHER_PROBABILITY
        if is_probability or is_teacher_probability:
            return expit(expected_values)
    return expected_values



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines the tabular explainer meta-api for returning the best explanation result based on the given model."""

from .common.base_explainer import BaseExplainer
from .common.constants import Defaults, ExplainParams, Extension, ModelTask
from .common.structured_model_explainer import PureStructuredModelExplainer
from .dataset.decorator import tabular_decorator
from .shap.deep_explainer import DeepExplainer
from .shap.gpu_kernel_explainer import GPUKernelExplainer
from .shap.kernel_explainer import KernelExplainer
from .shap.linear_explainer import LinearExplainer
from .shap.tree_explainer import TreeExplainer

InvalidExplainerErr = 'Could not find valid explainer to explain model'


def _get_uninitialized_explainers(use_gpu):
    """Return the uninitialized explainers used by the tabular explainer.

    :return: A list of the uninitialized explainers.
    :rtype: list
    """
    if use_gpu:
        return [GPUKernelExplainer]
    else:
        return [TreeExplainer, DeepExplainer, LinearExplainer, KernelExplainer]


class TabularExplainer(BaseExplainer):
    available_explanations = [Extension.GLOBAL, Extension.LOCAL]
    explainer_type = Extension.BLACKBOX

    """The tabular explainer meta-api for returning the best explanation result based on the given model.

    :param model: The model or pipeline to explain.
        A model that implements sklearn.predict() or sklearn.predict_proba() or pipeline function that accepts
        a 2d ndarray
    :type model: object
    :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
        initializing the explainer.
    :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
    :param explain_subset: List of feature indices. If specified, only selects a subset of the
        features in the evaluation dataset for explanation, which will speed up the explanation
        process when number of features is large and the user already knows the set of interested
        features. The subset can be the top-k features from the model summary. This argument is not supported when
        transformations are set.
    :type explain_subset: list[int]
    :param features: A list of feature names.
    :type features: list[str]
    :param classes: Class names as a list of strings. The order of the class names should match
        that of the model output.  Only required if explaining classifier.
    :type classes: list[str]
    :param transformations: sklearn.compose.ColumnTransformer or a list of tuples describing the column name and
        transformer. When transformations are provided, explanations are of the features before the transformation.
        The format for a list of transformations is same as the one here:
        https://github.com/scikit-learn-contrib/sklearn-pandas.

        If the user is using a transformation that is not in the list of sklearn.preprocessing transformations
        that are supported by the `interpret-community <https://github.com/interpretml/interpret-community>`_
        package, then this parameter cannot take a list of more than one column as input for the transformation.
        A user can use the following sklearn.preprocessing  transformations with a list of columns since these are
        already one to many or one to one: Binarizer, KBinsDiscretizer, KernelCenterer, LabelEncoder, MaxAbsScaler,
        MinMaxScaler, Normalizer, OneHotEncoder, OrdinalEncoder, PowerTransformer, QuantileTransformer,
        RobustScaler, StandardScaler.

        Examples for transformations that work::

            [
                (["col1", "col2"], sklearn_one_hot_encoder),
                (["col3"], None) #col3 passes as is
            ]
            [
                (["col1"], my_own_transformer),
                (["col2"], my_own_transformer),
            ]

        An example of a transformation that would raise an error since it cannot be interpreted as one to many::

            [
                (["col1", "col2"], my_own_transformer)
            ]

        The last example would not work since the interpret-community package can't determine whether
        my_own_transformer gives a many to many or one to many mapping when taking a sequence of columns.
    :type transformations: sklearn.compose.ColumnTransformer or list[tuple]
    :param allow_all_transformations: Allow many to many and many to one transformations
    :type allow_all_transformations: bool
    """

    def __init__(self, model, initialization_examples, explain_subset=None, features=None, classes=None,
                 transformations=None, allow_all_transformations=False, model_task=ModelTask.Unknown,
                 use_gpu=False, **kwargs):
        """Initialize the TabularExplainer.

        :param model: The model or pipeline to explain.
            A model that implements sklearn.predict() or sklearn.predict_proba() or pipeline function that accepts
            a 2d ndarray
        :type model: object
        :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
            initializing the explainer.
        :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param explain_subset: List of feature indices. If specified, only selects a subset of the
            features in the evaluation dataset for explanation, which will speed up the explanation
            process when number of features is large and the user already knows the set of interested
            features. The subset can be the top-k features from the model summary. This argument is not supported when
            transformations are set.
        :type explain_subset: list[int]
        :param features: A list of feature names.
        :type features: list[str]
        :param classes: Class names as a list of strings. The order of the class names should match
            that of the model output.  Only required if explaining classifier.
        :type classes: list[str]
        :param transformations: sklearn.compose.ColumnTransformer or a list of tuples describing the column name and
            transformer. When transformations are provided, explanations are of the features before the transformation.
            The format for a list of transformations is same as the one here:
            https://github.com/scikit-learn-contrib/sklearn-pandas.

            If the user is using a transformation that is not in the list of sklearn.preprocessing transformations
            that are supported by the `interpret-community <https://github.com/interpretml/interpret-community>`_
            package, then this parameter cannot take a list of more than one column as input for the transformation.
            A user can use the following sklearn.preprocessing  transformations with a list of columns since these are
            already one to many or one to one: Binarizer, KBinsDiscretizer, KernelCenterer, LabelEncoder, MaxAbsScaler,
            MinMaxScaler, Normalizer, OneHotEncoder, OrdinalEncoder, PowerTransformer, QuantileTransformer,
            RobustScaler, StandardScaler.

            Examples for transformations that work::

                [
                    (["col1", "col2"], sklearn_one_hot_encoder),
                    (["col3"], None) #col3 passes as is
                ]
                [
                    (["col1"], my_own_transformer),
                    (["col2"], my_own_transformer),
                ]

            An example of a transformation that would raise an error since it cannot be interpreted as one to many::

                [
                    (["col1", "col2"], my_own_transformer)
                ]

            The last example would not work since the interpret-community package can't determine whether
            my_own_transformer gives a many to many or one to many mapping when taking a sequence of columns.
        :type transformations: sklearn.compose.ColumnTransformer or list[tuple]
        :param allow_all_transformations: Allow many to many and many to one transformations
        :type allow_all_transformations: bool
        :param model_task: Optional parameter to specify whether the model is a classification or regression model.
            In most cases, the type of the model can be inferred based on the shape of the output, where a classifier
            has a predict_proba method and outputs a 2 dimensional array, while a regressor has a predict method and
            outputs a 1 dimensional array.
        :type model_task: str
        """
        super(TabularExplainer, self).__init__(**kwargs)
        self._logger.debug('Initializing TabularExplainer')

        if transformations is not None and explain_subset is not None:
            raise ValueError("explain_subset not supported with non-None transformations")

        self.model = model
        self.features = features
        self.classes = classes
        self.explain_subset = explain_subset
        self.transformations = transformations
        kwargs[ExplainParams.EXPLAIN_SUBSET] = self.explain_subset
        kwargs[ExplainParams.FEATURES] = features
        kwargs[ExplainParams.CLASSES] = classes
        uninitialized_explainers = _get_uninitialized_explainers(use_gpu)
        is_valid = False
        last_exception = None

        for uninitialized_explainer in uninitialized_explainers:
            try:
                if issubclass(uninitialized_explainer, PureStructuredModelExplainer):
                    self.explainer = uninitialized_explainer(
                        self.model, transformations=transformations,
                        allow_all_transformations=allow_all_transformations, **kwargs)
                else:
                    # Note: Unlike DeepExplainer, LinearExplainer does not need model_task
                    if uninitialized_explainer != LinearExplainer:
                        kwargs[ExplainParams.MODEL_TASK] = model_task
                    else:
                        kwargs.pop(ExplainParams.MODEL_TASK, None)
                    self.explainer = uninitialized_explainer(
                        self.model, initialization_examples, transformations=transformations,
                        allow_all_transformations=allow_all_transformations,
                        **kwargs)
                self._method = self.explainer._method
                self._logger.info('Initialized valid explainer {} with args {}'.format(self.explainer, kwargs))
                is_valid = True
                break
            except Exception as ex:
                last_exception = ex
                self._logger.info('Failed to initialize explainer {} due to error: {}'
                                  .format(uninitialized_explainer, ex))
        if not is_valid:
            self._logger.info(InvalidExplainerErr)
            raise ValueError(InvalidExplainerErr) from last_exception

    @tabular_decorator
    def explain_global(self, evaluation_examples, sampling_policy=None, include_local=True,
                       batch_size=Defaults.DEFAULT_BATCH_SIZE):
        """Globally explains the black box model or function.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param sampling_policy: Optional policy for sampling the evaluation examples.  See documentation on
            SamplingPolicy for more information.
        :type sampling_policy: interpret_community.common.policy.SamplingPolicy
        :param include_local: Include the local explanations in the returned global explanation.
            If include_local is False, will stream the local explanations to aggregate to global.
        :type include_local: bool
        :param batch_size: If include_local is False, specifies the batch size for aggregating
            local explanations to global.
        :type batch_size: int
        :return: A model explanation object. It is guaranteed to be a GlobalExplanation. If SHAP is used for the
            explanation, it will also have the properties of a LocalExplanation and the ExpectedValuesMixin. If the
            model does classification, it will have the properties of the PerClassMixin.
        :rtype: DynamicGlobalExplanation
        """
        kwargs = {ExplainParams.SAMPLING_POLICY: sampling_policy,
                  ExplainParams.INCLUDE_LOCAL: include_local,
                  ExplainParams.BATCH_SIZE: batch_size}
        return self.explainer.explain_global(evaluation_examples, **kwargs)

    @tabular_decorator
    def explain_local(self, evaluation_examples):
        """Locally explains the black box model or function.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: A model explanation object. It is guaranteed to be a LocalExplanation. If SHAP is used for the
            explanation, it will also have the properties of the ExpectedValuesMixin. If the model does
            classification, it will have the properties of the ClassesMixin.
        :rtype: DynamicLocalExplanation
        """
        return self.explainer.explain_local(evaluation_examples)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines helpful utilities for summarizing and uploading data."""

import logging

import numpy as np
from scipy.sparse import csr_matrix, eye, issparse
from scipy.sparse import vstack as sparse_vstack
from sklearn.preprocessing import normalize
from sklearn.utils import shuffle
from sklearn.utils.sparsefuncs import csc_median_axis_0

from .constants import Scipy
from .gpu_kmeans import kmeans
from .warnings_suppressor import shap_warnings_suppressor

with shap_warnings_suppressor():
    import shap

module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)

_VALUES = 'values'
_RANKING = 'ranking'
_FEATURES = 'features'


def _summarize_data(X, k=10, use_gpu=False, to_round_values=True):
    """Summarize a dataset.

    For dense dataset, use k mean samples weighted by the number of data points they
    each represent.
    For sparse dataset, use a sparse row for the background with calculated
    median for dense columns.

    :param X: Matrix of data samples to summarize (# samples x # features).
    :type X: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
    :param k: Number of cluster centroids to use for approximation.
    :type k: int
    :param to_round_values: When using kmeans, for each element of every cluster centroid to match the nearest value
        from X in the corresponding dimension. This ensures discrete features
        always get a valid value.  Ignored for sparse data sample.
    :type to_round_values: bool
    :return: summarized numpy array or csr_matrix object.
    :rtype: numpy.array or scipy.sparse.csr_matrix or DenseData
    """
    is_sparse = issparse(X)
    if not str(type(X)).endswith(".DenseData'>"):
        if is_sparse:
            module_logger.debug('Creating sparse data summary as csr matrix')
            # calculate median of sparse background data
            median_dense = csc_median_axis_0(X.tocsc())
            return csr_matrix(median_dense)
        elif len(X) > 10 * k:
            module_logger.debug('Create dense data summary with k-means')
            # use kmeans to summarize the examples for initialization
            # if there are more than 10 x k of them
            if use_gpu:
                return kmeans(X, k, to_round_values)
            else:
                return shap.kmeans(X, k, to_round_values)
    return X


def _should_compress_sparse_matrix(matrix):
    """Returns whether to compress the matrix, which can be sparse or dense format depending on optimal storage.

    If more than a third of the values are non-zero in the sparse matrix, we convert it to dense format.

    :param matrix: The matrix to compress.
    :type matrix: scipy.sparse.csr_matrix or list[scipy.sparse.csr_matrix]
    :return: Whether the matrix should be compressed.
    :rtype: bool
    """
    nnz = 0
    num_cells = 0
    if isinstance(matrix, list):
        for class_matrix in matrix:
            nnz += class_matrix.nnz
            num_cells += class_matrix.shape[0] * class_matrix.shape[1]
    else:
        nnz += matrix.nnz
        num_cells += matrix.shape[0] * matrix.shape[1]
    return nnz > num_cells / 3


def _get_raw_feature_importances(importance_values, raw_to_output_feature_maps):
    """Return raw feature importance values.

    :param importance_values: The importance values computed for the dataset.
    :type importance_values: numpy.array or list[scipy.sparse.csr_matrix]
    :param raw_to_output_feature_maps: A list of feature maps from raw to generated feature.
    :type raw_to_output_feature_maps: list[Union[numpy.array, scipy.sparse.csr_matrix]]
    :return: Raw feature importance values.
    :rtype: numpy.array
    """
    if not isinstance(raw_to_output_feature_maps, list):
        raise ValueError("raw_to_output_feature_maps should be a list of feature maps")
    if any(map(lambda x: (x < 0).sum(), raw_to_output_feature_maps)):
        raise ValueError("Feature maps should only contain non-negative values")
    if any(map(lambda x: not _is_one_to_many(x), raw_to_output_feature_maps)):
        print("Many to one/many maps found in input")

    raw_to_output_map = raw_to_output_feature_maps[0]
    for fmap in raw_to_output_feature_maps[1:]:
        if _is_identity(fmap):
            continue
        raw_to_output_map = raw_to_output_map.dot(fmap)

    # normalize column wise
    raw_to_output_map = normalize(raw_to_output_map, norm='l1', axis=0)

    orig_single_dimensional_importances = False
    if isinstance(importance_values, list):
        raw_importances = []
        for class_matrix in importance_values:
            raw_matrix = class_matrix.dot(raw_to_output_map.T)
            raw_importances.append(raw_matrix)
        is_sparse_matrix = len(raw_importances) > 0 and issparse(raw_importances[0])
        if is_sparse_matrix:
            if _should_compress_sparse_matrix(raw_importances):
                for i in range(len(raw_importances)):
                    raw_importances[i] = raw_importances[i].toarray()
                raw_importances = np.array(raw_importances)
        else:
            raw_importances = np.array(raw_importances)
        return raw_importances

    if len(importance_values.shape) < 2:
        importance_values = importance_values.reshape(1, -1)
        orig_single_dimensional_importances = True

    if issparse(raw_to_output_map):
        if len(importance_values.shape) > 2:
            raw_importances = _multiply_sparse_matrix_3d_numpy_tensor(importance_values, raw_to_output_map.T)
        else:
            if _is_identity(raw_to_output_map):
                raw_importances = importance_values
            else:
                raw_importances = raw_to_output_map.dot(importance_values.T).T
            if issparse(raw_importances) and _should_compress_sparse_matrix(raw_importances):
                raw_importances = raw_importances.toarray()
    else:
        raw_importances = importance_values.dot(raw_to_output_map.T)
    return raw_importances.squeeze(0) if orig_single_dimensional_importances else raw_importances


def _is_identity(matrix):
    """Checks if the given sparse matrix is identity matrix.

    :param matrix: sparse matrix
    :type matrix: scipy.sparse.csr_matrix
    :return: True if the matrix is an identity matrix.
    :rtype: bool
    """
    if not issparse(matrix):
        return False
    sh0 = matrix.shape[0]
    sh1 = matrix.shape[1]
    return sh0 == sh1 and matrix.nnz == sh0 and (matrix - eye(sh0, format=Scipy.CSR_FORMAT)).nnz == 0


def _multiply_sparse_matrix_3d_numpy_tensor(np_tensor, sp_matrix):
    """Multiply sparse matrix with a 3 dimension numpy array.

    :param np_tensor: numpy tensor of 3 dimensions.
    :type np_tensor: numpy.array
    :param sp_matrix: sparse matrix
    :type sp_matrix: scipy.sparse.csr_matrix
    :return: The product of numpy array and sparse matrix.
    :rtype: numpy.array
    """
    if len(np_tensor.shape) != 3:
        raise ValueError("Expected matrix of dimension 3. Got dimension {}.".format(len(np_tensor.shape)))

    output = np.zeros((*np_tensor.shape[:-1], sp_matrix.shape[1]))
    for i in range(np_tensor.shape[0]):
        output[i] = sp_matrix.T.dot(np_tensor[i].T).T

    return output


def _is_one_to_many(feature_map):
    return ((feature_map > 0).sum(axis=0) >= 2).sum() == 0


def _get_dense_examples(examples):
    if issparse(examples):
        module_logger.debug('converting sparse examples to regular array')
        return examples.toarray()
    return examples


def _convert_to_list(shap_values):
    module_logger.debug('converting numpy array of list of numpy array to list')
    classification = isinstance(shap_values, list)
    if classification:
        # shap_values is a list of ndarrays
        shap_values_numpy_free = []
        for array in shap_values:
            shap_values_numpy_free.append(array.tolist())
        return shap_values_numpy_free
    else:
        # shap_values is a single ndarray
        return shap_values.tolist()


def _generate_augmented_data(x, max_num_of_augmentations=np.inf):
    """Augment x by appending x with itself shuffled columnwise many times.

    :param x: data that has to be augmented, array or sparse matrix of 2 dimensions
    :type x: numpy.array or scipy.sparse.csr_matrix
    :param max_augment_data_size: number of times we stack permuted x to augment.
    :type max_augment_data_size: int
    :return: augmented data with roughly number of rows that are equal to number of columns
    :rtype: numpy.array or scipy.sparse.csr_matrix
    """
    x_augmented = x
    vstack = sparse_vstack if issparse(x) else np.vstack
    for i in range(min(x.shape[1] // x.shape[0] - 1, max_num_of_augmentations)):
        x_permuted = shuffle(x.T, random_state=i).T
        x_augmented = vstack([x_augmented, x_permuted])

    return x_augmented


def _scale_tree_shap(shap_values, expected_values, prediction):
    """Scale the log odds shap values from TreeExplainer to be in terms of probability.

    Note this is just an approximation, since the logistic function is non-linear.
    The function is a modified version of the implementations posted on shap github issues:
    https://github.com/slundberg/shap/issues/29#issuecomment-374928027
    https://github.com/slundberg/shap/issues/29#issuecomment-408385378

    :param shap_values: The shap values to transform from log odds to probabilities.
    :type shap_values: numpy.array
    :param expected_values: The expected values as probabilities.
    :type expected_values: numpy.array
    :param prediction: The predicted probability from the teacher model.
    :type prediction: numpy.array
    :return: The transformed tree shap values as probabilities.
    :rtype: list or numpy.array
    """
    # In multiclass case, use expected values and predictions per class
    if isinstance(shap_values, list):
        values_to_convert = shap_values
        for idx, shap_values in enumerate(values_to_convert):
            values_to_convert[idx] = _scale_single_shap_matrix(shap_values, expected_values[idx], prediction[:, idx])
        return values_to_convert
    else:
        if len(prediction.shape) == 1:
            return _scale_single_shap_matrix(shap_values, expected_values, prediction)
        else:
            # Note: in binary classification with teacher probability this will be index 1,
            # but in binary classification with surrogate probability this will be index 0
            return _scale_single_shap_matrix(shap_values, expected_values, prediction[:, -1])


def _scale_single_shap_matrix(shap_values, expectation, prediction):
    """Scale a single class matrix of shap values to sum to the teacher model prediction.

    :param shap_values: The shap values of the mimic model.
    :type shap_values: numpy.array
    :param expected_values: The expected values as probabilities (base values).
    :type expected_values: numpy.array
    :param prediction: The predicted probability from the teacher model.
    :type prediction: numpy.array
    :return: The transformed tree shap values as probabilities.
    :rtype: list or numpy.array
    """
    mimic_prediction = np.sum(shap_values, axis=1)
    error = prediction - mimic_prediction - expectation
    absolute_error = np.abs(error)
    error_sign = np.sign(error)
    absolute_shap_vector = np.abs(shap_values)
    absolute_shap_magnitude = np.sum(absolute_shap_vector, axis=1)
    # We divide by one, when we know the numerator is 0
    safe_absolute_magnitude_denominator = np.where(absolute_shap_magnitude > 0., absolute_shap_magnitude, 1)
    flat_magnitude = np.multiply(np.maximum(absolute_error - absolute_shap_magnitude, 0.), error_sign)
    feature_dimension = shap_values.shape[1]
    proportional_correction = np.divide(
        np.multiply(
            np.multiply(
                error_sign,
                np.minimum(absolute_error, absolute_shap_magnitude))[:, None],
            absolute_shap_vector),
        safe_absolute_magnitude_denominator[:, None])
    flat_correction = np.full(shap_values.shape, np.divide(flat_magnitude[:, None], feature_dimension))
    return shap_values + proportional_correction + flat_correction


def _convert_single_instance_to_multi(instance_shap_values):
    """Convert a single shap values instance to multi-instance form.

    :param instance_shap_values: The shap values calculated for a new instance.
    :type instance_shap_values: numpy.array or list
    :return: The instance converted to multi-instance form.
    :rtype: numpy.array or list
    """
    classification = isinstance(instance_shap_values, list)
    if classification:
        shap_values = instance_shap_values
        for i in range(len(shap_values)):
            shap_values[i] = shap_values[i].reshape(1, -1)
    else:
        shap_values = instance_shap_values
    return shap_values


def _append_shap_values_instance(shap_values, instance_shap_values):
    """Append a single instance of shap values to an existing multi-instance shap values list or array.

    :param shap_values: The existing shap values array or list.
    :type shap_values: numpy.array or list
    :param instance_shap_values: The shap values calculated for a new instance.
    :type instance_shap_values: numpy.array or list
    :return: The instance appended to the existing shap values.
    :rtype: numpy.array or list
    """
    classification = isinstance(instance_shap_values, list)
    if classification:
        for i in range(len(shap_values)):
            cols_dim = len(instance_shap_values[i].shape)
            # col_dim can only be 1 or 2 here, depending on data passed to shap
            if cols_dim != 2:
                output_size = shap_values[i].shape[-1]
                tmp_shap_values_2d = instance_shap_values[i].reshape(1, output_size)
            else:
                tmp_shap_values_2d = instance_shap_values[i]
            # Append rows
            shap_values[i] = np.append(shap_values[i], tmp_shap_values_2d, axis=0)
    else:
        shap_values = np.append(shap_values, instance_shap_values, axis=0)
    return shap_values


def _fix_linear_explainer_shap_values(model, shap_values):
    """Update the shap values from LinearExplainer in case the shape does not conform to API standards.

    :param model: The linear model.
    :type model: tuple of (coefficients, intercept) or sklearn.linear.* model
    :param shap_values: The existing shap values array or list.
    :type shap_values: numpy.array or list
    :return: The shap values reshaped into the correct form for given linear model.
    :rtype: numpy.array or list
    """
    # Temporary fix for a bug in shap for regression models
    if (type(model) == tuple and len(model) == 2):
        coef = model[0]
    elif hasattr(model, 'coef_') and hasattr(model, 'intercept_'):
        coef = model.coef_
    else:
        raise Exception('An unknown model type was passed: ' + str(type(model)))
    coef_shape_len = len(np.array(coef).shape)
    shap_values_shape_len = len(np.array(shap_values).shape)
    if coef_shape_len == 1 and shap_values_shape_len == 3:
        shap_values = shap_values[0]
    elif coef_shape_len == 2 and shap_values_shape_len == 2:
        shap_values = [-shap_values, shap_values]
    return shap_values


def _sort_values(values, order):
    return np.array(values)[order]


def _unsort_1d(values, order):
    """Unsort a sorted 1d array based on the order that was used to sort it.

    :param values: The array that has been sorted.
    :type values: numpy.array
    :param order: The order list that was originally used to sort values.
    :type order: numpy.array
    :return: The unsorted array.
    :rtype: numpy.array
    """
    order_inverse = [None] * len(order)
    for i in range(len(order)):
        order_inverse[order[i]] = i
    return np.array(values)[order_inverse]


def _order_imp(summary):
    """Compute the ranking of feature importance values.

    :param summary: A 3D array of the feature importance values to be ranked.
    :type summary: numpy.ndarray
    :return: The rank of the feature importance values.
    :rtype: numpy.ndarray
    """
    return summary.argsort()[..., ::-1]


def _sparse_order_imp_csr_matrix(local_importance_values, values_type=_RANKING, features=None, top_k=None):
    """Compute the ranking, names or values for sparse feature importance values on a single csr_matrix.

    :param local_importance_values: The local importance values to compute the ranking for.
    :type local_importance_values: scipy.sparse.csr_matrix
    :param values_type: The type of values, can be 'ranking', which returns the sorted
        indices, 'values', which returns the sorted values, or 'features', which returns
        the feature names.
    :type values_type: str
    :param features: The feature names.
    :type features: list[str]
    :param top_k: If specified, only the top k values will be returned.
    :type top_k: int
    :return: The rank of the non-zero sparse feature importance values.
    :rtype: list
    """
    sparse_ranking = []
    data = local_importance_values.data
    indices = local_importance_values.indices
    indptr = local_importance_values.indptr
    for i in range(len(indptr) - 1):
        rowstart = indptr[i]
        rowend = indptr[i + 1]
        row_values = data[rowstart:rowend]
        # compute the ranking of the values for each row
        values_ranking = row_values.argsort()[..., ::-1]
        if values_type == _VALUES:
            sorted_values = row_values[values_ranking]
            if top_k is not None:
                sorted_values = sorted_values[:top_k]
            sparse_ranking.append(sorted_values.tolist())
        elif values_type == _FEATURES:
            row_indices = indices[rowstart:rowend]
            indices_ranking = row_indices[values_ranking]
            if features is not None:
                ranked_features = np.array(features)[indices_ranking]
                if top_k is not None:
                    ranked_features = ranked_features[:top_k]
                sparse_ranking.append(ranked_features.tolist())
            else:
                if top_k is not None:
                    indices_ranking = indices_ranking[:top_k]
                sparse_ranking.append(indices_ranking.tolist())
        elif values_type == _RANKING:
            row_indices = indices[rowstart:rowend]
            # re-shuffle indices based on the ranking values
            indices_ranking = row_indices[values_ranking]
            if top_k is not None:
                indices_ranking = indices_ranking[:top_k]
            sparse_ranking.append(indices_ranking.tolist())
        else:
            raise ValueError("Unknown values_type specified: {}.".format(values_type))
    return sparse_ranking


def _sparse_order_imp(local_importance_values, values_type=_RANKING, features=None, top_k=None):
    """Compute the ranking for sparse feature importance values.

    :param local_importance_values: The local importance values to compute the ranking for.
    :type local_importance_values: scipy.sparse.csr_matrix or list[scipy.sparse.csr_matrix]
    :param values_type: The type of values, can be 'ranking', which returns the sorted
        indices, 'values', which returns the sorted values, or 'features', which returns
        the feature names.
    :type values_type: str
    :param features: The feature names.
    :type features: list[str]
    :param top_k: If specified, only the top k values will be returned.
    :type top_k: int
    :return: The rank of the non-zero sparse feature importance values.
    :rtype: list
    """
    if isinstance(local_importance_values, list):
        per_class_sparse_ranking = []
        for class_importance_values in local_importance_values:
            per_class_sparse_ranking.append(_sparse_order_imp_csr_matrix(class_importance_values,
                                                                         values_type=values_type,
                                                                         features=features,
                                                                         top_k=top_k))
        return per_class_sparse_ranking
    else:
        return _sparse_order_imp_csr_matrix(local_importance_values,
                                            values_type=values_type,
                                            features=features,
                                            top_k=top_k)


# sorts a single dimensional feature list according to order
def _sort_feature_list_single(features, order):
    return list(map(lambda x: features[x], order))


# returns a list of lists, where each internal list is the feature list sorted according to the order of a single class
def _sort_feature_list_multiclass(features, order):
    if hasattr(features, 'tolist'):
        features = features.tolist()
    return [list(map(lambda x: features[x], order_i)) for order_i in order]


# do the equivalent of a numpy array slice on a two-dimensional list
def _two_dimensional_slice(lst, end_index):
    return list(map(lambda x: x[:end_index], lst))


def _get_feature_map_from_list_of_indexes(indices_list):
    """Compute feature map from a list of indices from raw features to generated feature indices.

    :param indices_list: A list of lists of generated feature indices for each raw feature.
    :type indices_list: list[list[int]]
    :return: A feature map from raw to generated.
    :rtype: numpy.array
    """
    # sets for each list of generated features
    raw_to_gen = []
    for generated_index_list in indices_list:
        raw_to_gen.append(set(generated_index_list))

    num_gen_feats = 1 + max([max(x) for x in indices_list])
    num_raw = len(indices_list)
    # compute number of parents for a generated feature
    weights = np.zeros((num_gen_feats))
    for i in range(num_gen_feats):
        gen_feat_num_parents = sum([1 for gen_set in raw_to_gen if i in gen_set])
        if gen_feat_num_parents:
            # divide a generated feature weight equally amongst the raw features.
            weights[i] = 1.0 / gen_feat_num_parents

    feature_map = np.zeros((num_raw, num_gen_feats))
    for i in range(len(indices_list)):
        for j in indices_list[i]:
            feature_map[i, j] = weights[j]

    return feature_map



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines an explainer for DNN models."""

import logging
import sys

import numpy as np
from interpret_community._internal.raw_explain.raw_explain_utils import (
    get_datamapper_and_transformed_data, transform_with_datamapper)
from interpret_community.common.constants import (Attributes, Defaults,
                                                  DNNFramework, ExplainParams,
                                                  ExplainType, Extension,
                                                  ModelTask)

from ..common.aggregate import (add_explain_global_method,
                                init_aggregator_decorator)
from ..common.explanation_utils import _convert_to_list, _get_dense_examples
from ..common.structured_model_explainer import StructuredInitModelExplainer
from ..common.warnings_suppressor import shap_warnings_suppressor
from ..dataset.decorator import init_tabular_decorator, tabular_decorator
from ..explanation.explanation import (
    _create_local_explanation, _create_raw_feats_local_explanation,
    _get_raw_explainer_create_explanation_kwargs)
from .kwargs_utils import _get_explain_global_kwargs

with shap_warnings_suppressor():
    import shap


module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


try:
    import torch
except ImportError:
    module_logger.debug('Could not import torch, required if using a pytorch model')


class logger_redirector(object):
    """A redirector for system error output to logger."""

    def __init__(self, module_logger):
        """Initialize the logger_redirector.

        :param module_logger: The logger to use for redirection.
        :type module_logger: logger
        """
        self.logger = module_logger
        self.propagate = self.logger.propagate
        self.logger.propagate = False

    def __enter__(self):
        """Start the redirection for logging."""
        self.logger.debug("Redirecting user output to logger")
        self.original_stderr = sys.stderr
        sys.stderr = self

    def write(self, data):
        """Write the given data to logger.

        :param data: The data to write to logger.
        :type data: str
        """
        self.logger.debug(data)

    def flush(self):
        pass

    def close(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Finishes the redirection for logging."""
        try:
            if exc_val:
                # The default traceback.print_exc() expects a file-like object which
                # OutputCollector is not. Instead manually print the exception details
                # to the wrapped sys.stderr by using an intermediate string.
                # trace = traceback.format_tb(exc_tb)
                import traceback
                trace = "".join(traceback.format_exception(exc_type, exc_val, exc_tb))
                print(trace, file=sys.stderr)
        finally:
            sys.stderr = self.original_stderr
            self.logger.debug("User scope execution complete.")
            self.logger.propagate = self.propagate


def _get_dnn_model_framework(model):
    """Get the DNN model framework, taken from SHAP DeepExplainer.

    TODO: Refactor out SHAP's code so we can reference this method directly from SHAP.

    :return: The DNN Framework, PyTorch or TensorFlow.
    :rtype: str
    """
    actual_model = model[0] if type(model) is tuple else model
    return DNNFramework.PYTORCH if hasattr(actual_model, "named_parameters") else DNNFramework.TENSORFLOW


def _get_summary_data(initialization_examples, nclusters, framework):
    """Compute the summary data from the initialization examples.

    :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
        initializing the explainer.
    :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
    :param nclusters: Number of means to use for approximation. A dataset is summarized with nclusters mean
        samples weighted by the number of data points they each represent. When the number of initialization
        examples is larger than (10 x nclusters), those examples will be summarized with k-means where
        k = nclusters.
    :type nclusters: int
    :param framework: The framework, pytorch or tensorflow, for underlying DNN model.
    :type framework: str
    :return: A summarized matrix of feature vector examples (# examples x # features)
        for initializing the explainer.
    :rtype: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix or torch.autograd.Variable
    """
    initialization_examples.compute_summary(nclusters=nclusters)
    summary = initialization_examples.dataset
    summary_data = summary.data
    if framework == DNNFramework.PYTORCH:
        summary_data = torch.Tensor(summary_data)
    return summary_data


@add_explain_global_method
class DeepExplainer(StructuredInitModelExplainer):
    available_explanations = [Extension.GLOBAL, Extension.LOCAL]
    explainer_type = Extension.GREYBOX

    """An explainer for DNN models, implemented using shap's DeepExplainer, supports TensorFlow and PyTorch.

    :param model: The DNN model to explain.
    :type model: PyTorch or TensorFlow model
    :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
        initializing the explainer.
    :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
    :param explain_subset: List of feature indices. If specified, only selects a subset of the
        features in the evaluation dataset for explanation. The subset can be the top-k features
        from the model summary.
    :type explain_subset: list[int]
    :param nclusters: Number of means to use for approximation. A dataset is summarized with nclusters mean
        samples weighted by the number of data points they each represent. When the number of initialization
        examples is larger than (10 x nclusters), those examples will be summarized with k-means where
        k = nclusters.
    :type nclusters: int
    :param features: A list of feature names.
    :type features: list[str]
    :param classes: Class names as a list of strings. The order of the class names should match
        that of the model output.  Only required if explaining classifier.
    :type classes: list[str]
    :param transformations: sklearn.compose.ColumnTransformer or a list of tuples describing the column name and
        transformer. When transformations are provided, explanations are of the features before the transformation.
        The format for a list of transformations is same as the one here:
        https://github.com/scikit-learn-contrib/sklearn-pandas.

        If you are using a transformation that is not in the list of sklearn.preprocessing transformations that
        are supported by the `interpret-community <https://github.com/interpretml/interpret-community>`_
        package, then this parameter cannot take a list of more than one column as input for the transformation.
        You can use the following sklearn.preprocessing  transformations with a list of columns since these are
        already one to many or one to one: Binarizer, KBinsDiscretizer, KernelCenterer, LabelEncoder, MaxAbsScaler,
        MinMaxScaler, Normalizer, OneHotEncoder, OrdinalEncoder, PowerTransformer, QuantileTransformer,
        RobustScaler, StandardScaler.

        Examples for transformations that work::

            [
                (["col1", "col2"], sklearn_one_hot_encoder),
                (["col3"], None) #col3 passes as is
            ]
            [
                (["col1"], my_own_transformer),
                (["col2"], my_own_transformer),
            ]

        An example of a transformation that would raise an error since it cannot be interpreted as one to many::

            [
                (["col1", "col2"], my_own_transformer)
            ]

        The last example would not work since the interpret-community package can't determine whether
        my_own_transformer gives a many to many or one to many mapping when taking a sequence of columns.
    :type transformations: sklearn.compose.ColumnTransformer or list[tuple]
    :param allow_all_transformations: Allow many to many and many to one transformations
    :type allow_all_transformations: bool
    :param model_task: Optional parameter to specify whether the model is a classification or regression model.
    :type model_task: str
    """

    @init_tabular_decorator
    @init_aggregator_decorator
    def __init__(self, model, initialization_examples, explain_subset=None, nclusters=10,
                 features=None, classes=None, transformations=None, allow_all_transformations=False,
                 model_task=ModelTask.Unknown, is_classifier=None, **kwargs):
        """Initialize the DeepExplainer.

        :param model: The DNN model to explain.
        :type model: PyTorch or TensorFlow model
        :param initialization_examples: A matrix of feature vector examples (# examples x # features) for
            initializing the explainer.
        :type initialization_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param explain_subset: List of feature indices. If specified, only selects a subset of the
            features in the evaluation dataset for explanation. The subset can be the top-k features
            from the model summary.
        :type explain_subset: list[int]
        :param nclusters: Number of means to use for approximation. A dataset is summarized with nclusters mean
            samples weighted by the number of data points they each represent. When the number of initialization
            examples is larger than (10 x nclusters), those examples will be summarized with k-means where
            k = nclusters.
        :type nclusters: int
        :param features: A list of feature names.
        :type features: list[str]
        :param classes: Class names as a list of strings. The order of the class names should match
            that of the model output.  Only required if explaining classifier.
        :type classes: list[str]
        :param transformations: sklearn.compose.ColumnTransformer or a list of tuples describing the column name and
            transformer. When transformations are provided, explanations are of the features before the transformation.
            The format for a list of transformations is same as the one here:
            https://github.com/scikit-learn-contrib/sklearn-pandas.

            If you are using a transformation that is not in the list of sklearn.preprocessing transformations that
            are supported by the `interpret-community <https://github.com/interpretml/interpret-community>`_
            package, then this parameter cannot take a list of more than one column as input for the transformation.
            You can use the following sklearn.preprocessing  transformations with a list of columns since these are
            already one to many or one to one: Binarizer, KBinsDiscretizer, KernelCenterer, LabelEncoder, MaxAbsScaler,
            MinMaxScaler, Normalizer, OneHotEncoder, OrdinalEncoder, PowerTransformer, QuantileTransformer,
            RobustScaler, StandardScaler.

            Examples for transformations that work::

                [
                    (["col1", "col2"], sklearn_one_hot_encoder),
                    (["col3"], None) #col3 passes as is
                ]
                [
                    (["col1"], my_own_transformer),
                    (["col2"], my_own_transformer),
                ]

            An example of a transformation that would raise an error since it cannot be interpreted as one to many::

                [
                    (["col1", "col2"], my_own_transformer)
                ]

            The last example would not work since the interpret-community package can't determine whether
            my_own_transformer gives a many to many or one to many mapping when taking a sequence of columns.
        :type transformations: sklearn.compose.ColumnTransformer or list[tuple]
        :param allow_all_transformations: Allow many to many and many to one transformations
        :type allow_all_transformations: bool
        :param is_classifier: Optional parameter to specify whether the model is a classification or regression model.
            In most cases, the type of the model can be inferred based on the shape of the output, where a classifier
            has a predict_proba method and outputs a 2 dimensional array, while a regressor has a predict method and
            outputs a 1 dimensional array.
        :type is_classifier: bool
        :param model_task: Optional parameter to specify whether the model is a classification or regression model.
        :type model_task: str
        """
        self._datamapper = None
        if transformations is not None:
            self._datamapper, initialization_examples = get_datamapper_and_transformed_data(
                examples=initialization_examples, transformations=transformations,
                allow_all_transformations=allow_all_transformations)

        super(DeepExplainer, self).__init__(model, initialization_examples, **kwargs)
        self._logger.debug('Initializing DeepExplainer')
        self._method = 'shap.deep'
        self.features = features
        self.classes = classes
        self.nclusters = nclusters
        self.explain_subset = explain_subset
        self.transformations = transformations
        self.model_task = model_task
        self.framework = _get_dnn_model_framework(self.model)
        summary = _get_summary_data(self.initialization_examples, nclusters, self.framework)
        # Suppress warning message from Keras
        with logger_redirector(self._logger):
            self.explainer = shap.DeepExplainer(self.model, summary)

    @tabular_decorator
    def explain_global(self, evaluation_examples, sampling_policy=None,
                       include_local=True, batch_size=Defaults.DEFAULT_BATCH_SIZE):
        """Explain the model globally by aggregating local explanations to global.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param sampling_policy: Optional policy for sampling the evaluation examples.  See documentation on
            SamplingPolicy for more information.
        :type sampling_policy: interpret_community.common.policy.SamplingPolicy
        :param include_local: Include the local explanations in the returned global explanation.
            If include_local is False, will stream the local explanations to aggregate to global.
        :type include_local: bool
        :param batch_size: If include_local is False, specifies the batch size for aggregating
            local explanations to global.
        :type batch_size: int
        :return: A model explanation object. It is guaranteed to be a GlobalExplanation which also has the properties
            of LocalExplanation and ExpectedValuesMixin. If the model is a classifier, it will have the properties of
            PerClassMixin.
        :rtype: DynamicGlobalExplanation
        """
        kwargs = _get_explain_global_kwargs(sampling_policy, ExplainType.SHAP_DEEP, include_local, batch_size)
        kwargs[ExplainParams.INIT_DATA] = self.initialization_examples
        kwargs[ExplainParams.EVAL_DATA] = evaluation_examples
        kwargs[ExplainParams.NUM_FEATURES] = evaluation_examples.num_features
        return self._explain_global(evaluation_examples, **kwargs)

    def _get_explain_local_kwargs(self, evaluation_examples):
        """Get the kwargs for explain_local to create a local explanation.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: Args for explain_local.
        :rtype: dict
        """
        self._logger.debug('Explaining deep model')
        if self._datamapper is not None:
            evaluation_examples = transform_with_datamapper(evaluation_examples, self._datamapper)

        # sample the evaluation examples
        if self.sampling_policy is not None and self.sampling_policy.allow_eval_sampling:
            sampling_method = self.sampling_policy.sampling_method
            max_dim_clustering = self.sampling_policy.max_dim_clustering
            evaluation_examples.sample(max_dim_clustering, sampling_method=sampling_method)
        kwargs = {ExplainParams.METHOD: ExplainType.SHAP_DEEP}
        if self.classes is not None:
            kwargs[ExplainParams.CLASSES] = self.classes
        kwargs[ExplainParams.FEATURES] = evaluation_examples.get_features(features=self.features)
        kwargs[ExplainParams.NUM_FEATURES] = evaluation_examples.num_features
        evaluation_examples = evaluation_examples.dataset

        # for now convert evaluation examples to dense format if they are sparse
        # until DeepExplainer sparse support is added
        dense_examples = _get_dense_examples(evaluation_examples)
        if self.framework == DNNFramework.PYTORCH:
            dense_examples = torch.Tensor(dense_examples)
        shap_values = self.explainer.shap_values(dense_examples)
        # use model task to update structure of shap values
        single_output = isinstance(shap_values, list) and len(shap_values) == 1
        if single_output:
            if self.model_task == ModelTask.Regression:
                shap_values = shap_values[0]
            elif self.model_task == ModelTask.Classification:
                shap_values = [-shap_values[0], shap_values[0]]
        classification = isinstance(shap_values, list)
        if self.explain_subset:
            if classification:
                self._logger.debug('Classification explanation')
                for i in range(shap_values.shape[0]):
                    shap_values[i] = shap_values[i][:, self.explain_subset]
            else:
                self._logger.debug('Regression explanation')
                shap_values = shap_values[:, self.explain_subset]

        expected_values = None
        if hasattr(self.explainer, Attributes.EXPECTED_VALUE):
            self._logger.debug('reporting expected values')
            expected_values = self.explainer.expected_value
            if isinstance(expected_values, np.ndarray):
                expected_values = expected_values.tolist()
        else:
            return self._expected_values
        local_importance_values = _convert_to_list(shap_values)
        if classification:
            kwargs[ExplainParams.MODEL_TASK] = ExplainType.CLASSIFICATION
        else:
            kwargs[ExplainParams.MODEL_TASK] = ExplainType.REGRESSION
        kwargs[ExplainParams.MODEL_TYPE] = str(type(self.model))
        kwargs[ExplainParams.LOCAL_IMPORTANCE_VALUES] = np.array(local_importance_values)
        kwargs[ExplainParams.EXPECTED_VALUES] = np.array(expected_values)
        kwargs[ExplainParams.CLASSIFICATION] = classification
        kwargs[ExplainParams.INIT_DATA] = self.initialization_examples
        kwargs[ExplainParams.EVAL_DATA] = evaluation_examples
        return kwargs

    @tabular_decorator
    def explain_local(self, evaluation_examples):
        """Explain the model by using SHAP's deep explainer.

        :param evaluation_examples: A matrix of feature vector examples (# examples x # features) on which
            to explain the model's output.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: A model explanation object. It is guaranteed to be a LocalExplanation which also has the properties
            of ExpectedValuesMixin. If the model is a classifier, it will have the properties of the ClassesMixin.
        :rtype: DynamicLocalExplanation
        """
        kwargs = self._get_explain_local_kwargs(evaluation_examples)
        explanation = _create_local_explanation(**kwargs)

        if self._datamapper is None:
            return explanation
        else:
            # if transformations have been passed, then return raw features explanation
            raw_kwargs = _get_raw_explainer_create_explanation_kwargs(kwargs=kwargs)
            return _create_raw_feats_local_explanation(explanation, feature_maps=[self._datamapper.feature_map],
                                                       features=self.features, **raw_kwargs)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines the building blocks for explanations returned by explainers."""

from .serialization import load_explanation, save_explanation

__all__ = ['load_explanation', 'save_explanation']



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines metric constants for PFIExplainer."""

from enum import Enum


class MetricConstants(str, Enum):
    """The metric to use for PFIExplainer."""

    MEAN_ABSOLUTE_ERROR = 'mean_absolute_error'
    EXPLAINED_VARIANCE_SCORE = 'explained_variance_score'
    MEAN_SQUARED_ERROR = 'mean_squared_error'
    MEAN_SQUARED_LOG_ERROR = 'mean_squared_log_error'
    MEDIAN_ABSOLUTE_ERROR = 'median_absolute_error'
    R2_SCORE = 'r2_score'
    AVERAGE_PRECISION_SCORE = 'average_precision_score'
    F1_SCORE = 'f1_score'
    FBETA_SCORE = 'fbeta_score'
    PRECISION_SCORE = 'precision_score'
    RECALL_SCORE = 'recall_score'


# Note: Error metrics are those for which a higher score is worse, not better
# These should be given a negative value when computing the feature importance
error_metrics = {MetricConstants.MEAN_ABSOLUTE_ERROR, MetricConstants.MEAN_SQUARED_ERROR,
                 MetricConstants.MEAN_SQUARED_LOG_ERROR, MetricConstants.MEDIAN_ABSOLUTE_ERROR}



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Module for permutation feature importance."""

from .permutation_importance import PFIExplainer

__all__ = ['PFIExplainer']



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Suppresses warnings on imports."""

import os
import warnings

TF_CPP_MIN_LOG_LEVEL = 'TF_CPP_MIN_LOG_LEVEL'


class tf_warnings_suppressor(object):
    """Context manager to suppress warnings from tensorflow."""

    def __init__(self):
        """Initialize the tf_warnings_suppressor."""
        self._entered = False
        if TF_CPP_MIN_LOG_LEVEL in os.environ:
            self._default_tf_log_level = os.environ[TF_CPP_MIN_LOG_LEVEL]
        else:
            self._default_tf_log_level = '0'

    def __enter__(self):
        """Begins suppressing tensorflow warnings."""
        if self._entered:
            raise RuntimeError("Cannot enter %r twice" % self)
        self._entered = True
        os.environ[TF_CPP_MIN_LOG_LEVEL] = '2'

    def __exit__(self, *exc_info):
        """Finishes suppressing tensorflow warnings."""
        if not self._entered:
            raise RuntimeError("Cannot exit %r without entering first" % self)
        os.environ[TF_CPP_MIN_LOG_LEVEL] = self._default_tf_log_level


class shap_warnings_suppressor(object):
    """Context manager to suppress warnings from shap."""

    def __init__(self):
        """Initialize the shap_warnings_suppressor."""
        self._catch_warnings = warnings.catch_warnings()
        self._tf_warnings_suppressor = tf_warnings_suppressor()
        self._entered = False
        if TF_CPP_MIN_LOG_LEVEL in os.environ:
            self._default_tf_log_level = os.environ[TF_CPP_MIN_LOG_LEVEL]
        else:
            self._default_tf_log_level = '0'

    def __enter__(self):
        """Begins suppressing shap warnings."""
        if self._entered:
            raise RuntimeError("Cannot enter %r twice" % self)
        self._entered = True
        self._tf_warnings_suppressor.__enter__()
        log = self._catch_warnings.__enter__()
        warnings.filterwarnings('ignore', 'Starting from version 2.2.1', UserWarning)
        return log

    def __exit__(self, *exc_info):
        """Finishes suppressing shap warnings."""
        if not self._entered:
            raise RuntimeError("Cannot exit %r without entering first" % self)
        self._tf_warnings_suppressor.__exit__()
        self._catch_warnings.__exit__()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import numpy as np
import pandas
from scipy.sparse import csr_matrix, hstack, issparse
from sklearn.pipeline import Pipeline

from .data_mapper_utils import (
    extract_column, get_transformer_config_tuples_from_column_transformer,
    get_transformer_config_tuples_from_transformations_list)
from .feature_mappers import (FuncTransformer, IdentityMapper,
                              ManytoManyMapper, PassThroughMapper,
                              encoders_to_mappers_dict,
                              get_feature_mapper_for_pipeline)

try:
    from sklearn.compose import ColumnTransformer
    column_transformer_exists = True
except ImportError:
    column_transformer_exists = False


class TransformationsListTransformer:
    """Computes transformations on input data from a list of columns and pre-trained transformers in the same format
    that sklearn-pandas' DataFrameMapper expects."""
    def __init__(self, transformer_config_tuples):
        """

        :param transformer_config_tuples: transformer, transformer config associated with the transformer
        :type transformer_config_tuples: [(transformer, TransformerConfig)] where transformer is an object with a
        .transform method.
        """
        self._transformer_config_tuples = transformer_config_tuples

    def transform(self, x):
        """Transform data when the user provides a list of transformations in sklearn-pandas format.

        :param x: input data
        :type x: numpy.array or scipy.sparse matrix
        :return: transformed data
        """
        results = []
        for tr, tr_config in self._transformer_config_tuples:
            x_column = extract_column(x, tr_config)
            if tr is None:
                # The column is used as if transformer is None
                results.append(x_column)
            else:
                results.append(tr.transform(x_column))

        if any(map(issparse, results)):
            return hstack(results).tocsr()
        else:
            return np.hstack(tuple(results))


class DataMapper(object):
    """Class to transform raw features to engineered features using list of transformations."""

    def __init__(self, transformations=None, allow_all_transformations=False):
        """Initialize DataMapper object.

        :param transformations: List of (column_name, transformer) tuples or sklearn.compose.ColumnTransformer
        :type transformations: list[tuple(str, (class containing .transform method))]
        :param examples: DataFrame or numpy array of input
        :type examples: pandas.DataFrame or numpy.array
        :param allow_all_transformations: Allow many to many and many to one transformations
        :type allow_all_transformations: bool
        """
        self._transformer_config_tuples = self._get_transformer_config_tuples(transformations)
        self._feature_mappers_pipeline = self._build_feature_mappers_pipeline(
            self._transformer_config_tuples, allow_all_transformations=allow_all_transformations)
        self._transform = self._get_transform_func(transformations)
        self._feature_map = None

    @property
    def feature_map(self):
        """Feature map from raw to generated.

        :return: mapping from raw to generated features
        :rtype: numpy.array
        """
        if self._feature_map is None:
            raise ValueError("Feature map not built. Run transform first.")
        return self._feature_map

    def _get_transform_func(self, transformations):
        """Gets the transform function that produces transformed data.

        :param transformations: list of transformations in sklearn-pandas format or sklearn.compose.ColumnTransformer
        object
        :type transformations: list[tuple] or sklearn.compose.ColumnTransformer
        :return: function that maps input data to transformed data.
        """
        if isinstance(transformations, list):
            return TransformationsListTransformer(self._transformer_config_tuples).transform
        elif column_transformer_exists and isinstance(transformations, ColumnTransformer):
            return transformations.transform
        else:
            raise Exception("{} not supported as transformations argument".format(type(transformations)))

    def _get_transformer_config_tuples(self, transformations):
        """Get input transformations in the form of transformer, TransformerConfig tuples

        :param transformations: list of transformations in sklearn-pandas format or sklearn.compose.ColumnTransformer
        object
        :type transformations: list[tuple] or sklearn.compose.ColumnTransformer
        :return: list of tuples of the type (transformer, TransformerConfig)
        """

        if isinstance(transformations, list):
            return get_transformer_config_tuples_from_transformations_list(transformations)
        elif column_transformer_exists and isinstance(transformations, ColumnTransformer):
            return get_transformer_config_tuples_from_column_transformer(transformations)

        raise Exception("Non-supported transformations argument passed.")

    def _build_feature_mappers_pipeline(self, transformations, allow_all_transformations=False):
        """Generate a list of FeatureMappers that can transform as well as contain a featmap property

        :param transformations: from a list of transformer, TransformerConfig tuples, generate featuremappers.
        :type transformations: list[tuple]
        :param allow_all_transformations: Allow many to many and many to one transformations
        :type allow_all_transformations: bool
        :return: list of tuples of TransformerConfig and FeatureMapper
        :rtype: list[tuple(TransformerConfig, FeatureMapper)]
        """
        result = []
        for transformer, config in transformations:
            result.append((config, self._get_feature_mapper(transformer, config, allow_all_transformations)))

        return result

    def _get_feature_mapper(self, transformer, transformer_config, allow_all_transformations=False):
        """Get FeatureMapper from transformer and columns list that can also get the associated featmap.

        :param transformer: object that has a transform method
        :type transformer: class that has a transform method
        :param transformer_config: TransformerConfig needed to get the right shape of input data for transformer
        :type transformer_config: TransformerConfig
        :param allow_all_transformations: Allow many to many and many to one transformations
        :type allow_all_transformations: bool
        :return: feature mapper associated with the transformer
        :rtype: FeatureMapper
        """
        # return input as output if transformer is None
        if transformer is None:
            return IdentityMapper(FuncTransformer(lambda x: x))

        # if there is only one column, we can just look at the shape of final transformers and get result
        if len(transformer_config.columns) == 1:
            return PassThroughMapper(transformer)

        # if it is one of the supported transformations
        transformer_type = type(transformer)
        if transformer_type in encoders_to_mappers_dict:
            return encoders_to_mappers_dict[transformer_type](transformer)

        if isinstance(transformer, Pipeline):
            return get_feature_mapper_for_pipeline(transformer)

        # its a many to many or many to one map if we end up here
        if allow_all_transformations:
            return ManytoManyMapper(transformer)

        raise ValueError("Many to many or many to one transformers not supported in raw explanations when "
                         "explainer instantiated with allow_all_transformations is set to False. Change this "
                         "parameter to True in order to get explanations.")

    def _build_feature_map(self, x, columns):
        """Build the feature map from the feature maps of list of transformation wrappers in DataMapper.

        :param columns: input columns either as string names for dataframe or list of integers for numpy array
        :type columns: list[str] or list[int]
        """
        # run all transform methods in feature mappers so that we have column counts filled in
        self._run_feature_mappers_transform(x)

        num_output_columns = int(sum([x[1].feature_map.shape[1] for x in self._feature_mappers_pipeline]))
        feature_map = csr_matrix((len(columns), num_output_columns))
        raw_feats_indices = dict([(col, i) for i, col in enumerate(columns)])
        curr_output_cols = 0
        for transformer_config, feature_mapper in self._feature_mappers_pipeline:
            column_names = transformer_config.columns
            num_output_columns_for_transformer = feature_mapper.feature_map.shape[1]
            curr_start = curr_output_cols
            curr_end = curr_output_cols + num_output_columns_for_transformer
            for i, col in enumerate(column_names):
                feature_map[raw_feats_indices[col], curr_start:curr_end] = feature_mapper.feature_map[i]

            curr_output_cols += num_output_columns_for_transformer

        # Convert to dense format if more than half of the matrix has non-zeros
        if feature_map.shape[0] * feature_map.shape[1] / feature_map.getnnz() <= 2.0:
            feature_map = feature_map.toarray()

        self._feature_map = feature_map

    def _run_feature_mappers_transform(self, x):
        """Run the transform methods associated with each feature_mapper. This will set the featmaps.

        :param x: input data
        :type x: numpy.array or DataFrame
        :return: numpy.array or scipy.sparse.csr_matrix
        """
        for transformer_config, feature_mapper in self._feature_mappers_pipeline:
            feature_mapper.transform(extract_column(x, transformer_config))

    def transform(self, x):
        """Transform input data given the transformations.

        :param x: input data
        :type x: pandas.DataFrame or numpy.array
        :return: transformed data
        :rtype: numpy.array or scipy.sparse.csr_matrix
        """
        if self._feature_map is None:
            # pass a single example through the transformations list to build feature map
            columns = x.columns if isinstance(x, pandas.DataFrame) else list(range(x.shape[1]))
            self._build_feature_map(x[:1], columns)

        return self._transform(x)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines utilities for getting progress status for explanation."""


def _tqdm_func(*i, **kwargs):
    return i[0]


def get_tqdm(logger, show_progress):
    """Get the tqdm progress bar function.

    :param logger: The logger for logging info messages.
    :type logger: logger
    :param show_progress: Default to 'True'.  Determines whether to display the explanation status bar
        when using PFIExplainer.
    :type show_progress: bool
    :return: The tqdm (https://github.com/tqdm/tqdm) progress bar.
    :rtype: function
    """
    # This is used to get around code style build error F811
    tqdm = _tqdm_func
    if show_progress:
        try:
            from tqdm import tqdm
        except ImportError:
            logger.info('Failed to import tqdm to show progress bar')
    return tqdm



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines a light-weight chained identity for logging."""

import logging


class ChainedIdentity(object):
    """The base class for logging information."""

    def __init__(self, **kwargs):
        """Initialize the ChainedIdentity."""
        self._logger = logging.getLogger("interpret_community").getChild(self.__class__.__name__)
        self._identity = self.__class__.__name__
        super(ChainedIdentity, self).__init__(**kwargs)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Defines an explainable linear model."""

import numpy as np
from scipy.sparse import csr_matrix, issparse
from sklearn.linear_model import (Lasso, LinearRegression, LogisticRegression,
                                  SGDClassifier, SGDRegressor)

from ...common.constants import ExplainableModelType, Extension, SHAPDefaults
from ...common.explanation_utils import _summarize_data
from ...common.warnings_suppressor import shap_warnings_suppressor
from .explainable_model import (BaseExplainableModel, _clean_doc,
                                _get_initializer_args)

with shap_warnings_suppressor():
    import shap

DEFAULT_RANDOM_STATE = 123
FEATURE_DEPENDENCE = 'interventional'
LINEAR_PENALTY = 'penalty'
LINEAR_L2 = 'l2'
LINEAR_SOLVER = 'solver'
LINEAR_LBFGS = 'lbfgs'
LINEAR_MULTICLASS = 'multi_class'
LINEAR_MULTINOMIAL = 'multinomial'
LINEAR_RANDOM_STATE = 'random_state'


class LinearExplainer(shap.LinearExplainer):
    """Linear explainer with support for sparse data and sparse output."""

    def __init__(self, model, data, feature_dependence=FEATURE_DEPENDENCE):
        """Initialize the LinearExplainer.

        :param model: The linear model to compute the shap values for as a (coef, intercept) tuple.
        :type model: (numpy.matrix, double)
        :param data: The mean and covariance of the dataset.
        :type data: (scipy.csr_matrix, None)
        :param feature_dependence: Indicates the type of feature dependency, interventional by default.
        :type feature_dependence: str
        """
        self.is_sparse = data[1] is None
        if self.is_sparse:
            # Sparse case
            self.coef = model[0]
            if not issparse(self.coef):
                self.coef = np.asmatrix(self.coef)
            self.intercept = model[1]
            self._background = data[0]
            self.expected_value = np.array(self._background.dot(self.coef.T) + self.intercept).flatten()
        else:
            # Dense case
            super(LinearExplainer, self).__init__(model, data, feature_dependence=feature_dependence)

    def shap_values(self, evaluation_examples):
        """Estimate the SHAP values for a set of samples.

        :param evaluation_examples: The evaluation examples.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: For models with a single output this returns a matrix of SHAP values
            (# samples x # features). Each row sums to the difference between the model output for that
            sample and the expected value of the model output (which is stored as expected_value
            attribute of the explainer).
        :rtype: Union[list, numpy.ndarray]
        """
        if self.is_sparse:
            assert len(evaluation_examples.shape) == 2, "Sparse instance must have 2 dimensions!"
            assert self.coef.shape[0] == 1, "Multiclass coefficients need to be evaluated separately"
            mean_multiplier = csr_matrix(np.ones((evaluation_examples.shape[0], 1)))
            return (evaluation_examples - mean_multiplier * self._background).multiply(self.coef[0]).tocsr()
        else:
            return super().shap_values(evaluation_examples)


def _create_linear_explainer(model, multiclass, mean, covariance, seed):
    """Create the linear explainer or, in multiclass case, list of explainers.

    :param model: The linear model to compute the shap values for.
        A linear model that implements sklearn.predict or sklearn.predict_proba.
    :type model: object
    :param multiclass: True if this is a multiclass model.
    :type multiclass: bool
    :param mean: The mean of the dataset by columns.
    :type mean: numpy.array
    :param covariance: The covariance matrix of the dataset.
    :type covariance: numpy.array
    :param seed: Random number seed.
    :type seed: int
    """
    np.random.seed(seed)
    if multiclass:
        explainers = []
        coefs = model.coef_
        intercepts = model.intercept_
        if isinstance(intercepts, np.ndarray):
            intercepts = intercepts.tolist()
        if isinstance(intercepts, list):
            coef_intercept_list = zip(coefs, intercepts)
        else:
            coef_intercept_list = [(coef, intercepts) for coef in coefs]
        for class_coef, intercept in coef_intercept_list:
            linear_explainer = LinearExplainer((class_coef, intercept), (mean, covariance),
                                               feature_dependence=SHAPDefaults.INDEPENDENT)
            explainers.append(linear_explainer)
        return explainers
    else:
        model_coef = model.coef_
        model_intercept = model.intercept_
        return LinearExplainer((model_coef, model_intercept), (mean, covariance),
                               feature_dependence=SHAPDefaults.INDEPENDENT)


def _compute_local_shap_values(linear_explainer, evaluation_examples, classification):
    """Compute the local shap values.

    :param linear_explainer: The linear explainer or list of linear explainers in multiclass case.
    :type linear_explainer: Union[LinearExplainer, list[LinearExplainer]]
    :param evaluation_examples: The evaluation examples.
    :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
    """
    # Multiclass case
    if isinstance(linear_explainer, list):
        shap_values = []
        for explainer in linear_explainer:
            explainer_shap_values = explainer.shap_values(evaluation_examples)
            if isinstance(explainer_shap_values, list):
                explainer_shap_values = explainer_shap_values[0]
            shap_values.append(explainer_shap_values)
        return shap_values
    shap_values = linear_explainer.shap_values(evaluation_examples)
    if not classification and isinstance(shap_values, list):
        shap_values = shap_values[0]
    return shap_values


def _compute_background_data(dataset):
    """Compute the background data for linear explainer.

    :param dataset: The input dataset to compute background for.
    :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
    """
    background = _summarize_data(dataset)
    if str(type(background)).endswith(".DenseData'>"):
        background = background.data
    if not issparse(background) and len(background.shape) == 2:
        mean_shape = background.shape[1]
        # Take mean of clusters to get better representation of background
        if background.shape[0] > 1:
            background = background.mean(axis=0)
        # Check again prior to reshape
        if len(background.shape) == 2:
            background = background.reshape((mean_shape,))
    return background


class LinearExplainableModel(BaseExplainableModel):
    available_explanations = [Extension.GLOBAL, Extension.LOCAL]
    explainer_type = Extension.GLASSBOX

    """Linear explainable model.

    :param multiclass: Set to true to generate a multiclass model.
    :type multiclass: bool
    :param random_state: Int to seed the model.
    :type random_state: int
    :param classification: Indicates whether the model is used for classification or regression scenario.
    :type classification: bool
    :param sparse_data: Indicates whether the training data will be sparse.
    :type sparse_data: bool
    """

    def __init__(self, multiclass=False, random_state=DEFAULT_RANDOM_STATE, classification=True,
                 sparse_data=False, **kwargs):
        """Initialize the LinearExplainableModel.

        :param multiclass: Set to true to generate a multiclass model.
        :type multiclass: bool
        :param random_state: Int to seed the model.
        :type random_state: int
        :param classification: Indicates whether the model is used for classification or regression scenario.
        :type classification: bool
        :param sparse_data: Indicates whether the training data will be sparse.
        :type sparse_data: bool
        """
        self.multiclass = multiclass
        self.random_state = random_state
        self._sparse_data = sparse_data
        if self.multiclass:
            initializer = LogisticRegression
            if self._sparse_data:
                kwargs[LINEAR_PENALTY] = LINEAR_L2
            if LINEAR_SOLVER not in kwargs:
                kwargs[LINEAR_SOLVER] = LINEAR_LBFGS
                kwargs[LINEAR_MULTICLASS] = LINEAR_MULTINOMIAL
            kwargs[LINEAR_RANDOM_STATE] = random_state
        else:
            if self._sparse_data:
                initializer = Lasso
            else:
                initializer = LinearRegression
        initializer_args = _get_initializer_args(kwargs)
        self._linear = initializer(**initializer_args)
        super(LinearExplainableModel, self).__init__(**kwargs)
        self._logger.debug('Initializing LinearExplainableModel')
        self._method = 'linear'
        self._linear_explainer = None
        self._classification = classification

    __init__.__doc__ = (__init__.__doc__ +
                        '\nIf multiclass=True, uses the parameters for LogisticRegression:\n' +
                        _clean_doc(LogisticRegression.__doc__) +
                        '\nOtherwise, if multiclass=False, uses the parameters for LinearRegression:\n' +
                        _clean_doc(LinearRegression.__doc__))

    def fit(self, dataset, labels, **kwargs):
        """Call linear fit to fit the explainable model.

        Store the mean and covariance of the background data for local explanation.

        :param dataset: The dataset to train the model on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param labels: The labels to train the model on.
        :type labels: numpy.array
        """
        self._linear.fit(dataset, labels, **kwargs)
        self._background = _compute_background_data(dataset)
        if not issparse(dataset):
            self.covariance = np.cov(dataset, rowvar=False)
        else:
            # Not needed for sparse case
            self.covariance = None

    fit.__doc__ = (fit.__doc__ +
                   '\nIf multiclass=True, uses the parameters for LogisticRegression:\n' +
                   _clean_doc(LogisticRegression.fit.__doc__) +
                   '\nOtherwise, if multiclass=False, uses the parameters for LinearRegression:\n' +
                   _clean_doc(LinearRegression.fit.__doc__))

    def predict(self, dataset, **kwargs):
        """Call linear predict to predict labels using the explainable model.

        :param dataset: The dataset to predict on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The predictions of the model.
        :rtype: list
        """
        return self._linear.predict(dataset)

    predict.__doc__ = (predict.__doc__ +
                       '\nIf multiclass=True, uses the parameters for LogisticRegression:\n' +
                       _clean_doc(LogisticRegression.predict.__doc__) +
                       '\nOtherwise, if multiclass=False, uses the parameters for LinearRegression:\n' +
                       _clean_doc(LinearRegression.predict.__doc__))

    def predict_proba(self, dataset, **kwargs):
        """Call linear predict_proba to predict probabilities using the explainable model.

        :param dataset: The dataset to predict probabilities on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The predictions of the model.
        :rtype: list
        """
        if self.multiclass:
            return self._linear.predict_proba(dataset)
        else:
            raise Exception('predict_proba not supported for regression or binary classification dataset')

    predict_proba.__doc__ = (predict_proba.__doc__ +
                             '\nIf multiclass=True, uses the parameters for LogisticRegression:\n' +
                             _clean_doc(LogisticRegression.predict_proba.__doc__) +
                             '\nOtherwise predict_proba is not supported for regression or binary classification.\n')

    def explain_global(self, **kwargs):
        """Call coef to get the global feature importances from the linear surrogate model.

        :return: The global explanation of feature importances.
        :rtype: list
        """
        coef = self._linear.coef_
        if (len(coef.shape) == 2):
            return np.mean(coef, axis=0)
        return coef

    def explain_local(self, evaluation_examples, **kwargs):
        """Use LinearExplainer to get the local feature importances from the trained explainable model.

        :param evaluation_examples: The evaluation examples to compute local feature importances for.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The local explanation of feature importances.
        :rtype: Union[list, numpy.ndarray]
        """
        if self._linear_explainer is None:
            self._linear_explainer = _create_linear_explainer(self._linear, self.multiclass, self._background,
                                                              self.covariance, self.random_state)
        return _compute_local_shap_values(self._linear_explainer, evaluation_examples, self._classification)

    @property
    def expected_values(self):
        """Use LinearExplainer to get the expected values.

        :return: The expected values of the linear model.
        :rtype: list
        """
        if self._linear_explainer is None:
            self._linear_explainer = _create_linear_explainer(self._linear, self.multiclass, self._background,
                                                              self.covariance, self.random_state)
        if isinstance(self._linear_explainer, list):
            expected_values = []
            for explainer in self._linear_explainer:
                expected_values.append(explainer.expected_value)
            return expected_values
        else:
            expected_values = self._linear_explainer.expected_value
            if self._classification and not self.multiclass:
                expected_values = [-expected_values, expected_values]
            return expected_values

    @property
    def model(self):
        """Retrieve the underlying model.

        :return: The linear model, either classifier or regressor.
        :rtype: Union[LogisticRegression, LinearRegression]
        """
        return self._linear

    @staticmethod
    def explainable_model_type():
        """Retrieve the model type.

        :return: Linear explainable model type.
        :rtype: ExplainableModelType
        """
        return ExplainableModelType.LINEAR_EXPLAINABLE_MODEL_TYPE


class SGDExplainableModel(BaseExplainableModel):
    available_explanations = [Extension.GLOBAL, Extension.LOCAL]
    explainer_type = Extension.GLASSBOX

    """Stochastic Gradient Descent explainable model.

    :param multiclass: Set to true to generate a multiclass model.
    :type multiclass: bool
    :param random_state: Int to seed the model.
    :type random_state: int
    """

    def __init__(self, multiclass=False, random_state=DEFAULT_RANDOM_STATE, classification=True, **kwargs):
        """Initialize the SGDExplainableModel.

        :param multiclass: Set to true to generate a multiclass model.
        :type multiclass: bool
        :param random_state: Int to seed the model.
        :type random_state: int
        """
        self.multiclass = multiclass
        self.random_state = random_state
        if self.multiclass:
            initializer = SGDClassifier
        else:
            initializer = SGDRegressor
        initializer_args = _get_initializer_args(kwargs)
        self._sgd = initializer(random_state=random_state, **initializer_args)
        super(SGDExplainableModel, self).__init__(**kwargs)
        self._logger.debug('Initializing SGDExplainableModel')
        self._method = 'sgd'
        self._sgd_explainer = None
        self._classification = classification

    __init__.__doc__ = (__init__.__doc__ +
                        '\nIf multiclass=True, uses the parameters for SGDClassifier:\n' +
                        _clean_doc(SGDClassifier.__doc__) +
                        '\nOtherwise, if multiclass=False, uses the parameters for SGDRegressor:\n' +
                        _clean_doc(SGDRegressor.__doc__))

    def fit(self, dataset, labels, **kwargs):
        """Call linear fit to fit the explainable model.

        Store the mean and covariance of the background data for local explanation.

        :param dataset: The dataset to train the model on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :param labels: The labels to train the model on.
        :type labels: numpy.array
        """
        self._sgd.fit(dataset, labels, **kwargs)
        self._background = _compute_background_data(dataset)
        if not issparse(dataset):
            self.covariance = np.cov(dataset, rowvar=False)
        else:
            # Not needed for sparse case
            self.covariance = None

    fit.__doc__ = (fit.__doc__ +
                   '\nIf multiclass=True, uses the parameters for SGDClassifier:\n' +
                   _clean_doc(SGDClassifier.fit.__doc__) +
                   '\nOtherwise, if multiclass=False, uses the parameters for SGDRegressor:\n' +
                   _clean_doc(SGDRegressor.fit.__doc__))

    def predict(self, dataset, **kwargs):
        """Call SGD predict to predict labels using the explainable model.

        :param dataset: The dataset to predict on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The predictions of the model.
        :rtype: list
        """
        return self._sgd.predict(dataset)

    predict.__doc__ = (predict.__doc__ +
                       '\nIf multiclass=True, uses the parameters for SGDClassifier:\n' +
                       _clean_doc(SGDClassifier.predict.__doc__) +
                       '\nOtherwise, if multiclass=False, uses the parameters for SGDRegressor:\n' +
                       _clean_doc(SGDRegressor.predict.__doc__))

    def predict_proba(self, dataset, **kwargs):
        """Call SGD predict_proba to predict probabilities using the explainable model.

        :param dataset: The dataset to predict probabilities on.
        :type dataset: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The predictions of the model.
        :rtype: list
        """
        if self.multiclass:
            return self._sgd.predict_proba(dataset)
        else:
            raise Exception('predict_proba not supported for regression or binary classification dataset')

    predict_proba.__doc__ = (predict_proba.__doc__ +
                             '\nIf multiclass=True, uses the parameters for SGDClassifier:\n' +
                             _clean_doc(SGDClassifier.predict_proba.__doc__)
                             .replace(':class:`sklearn.calibration.CalibratedClassifierCV`',
                                      'CalibratedClassifierCV') +
                             '\nOtherwise predict_proba is not supported for regression or binary classification.\n')

    def explain_global(self, **kwargs):
        """Call coef to get the global feature importances from the SGD surrogate model.

        :return: The global explanation of feature importances.
        :rtype: list
        """
        coef = self._sgd.coef_
        if (len(coef.shape) == 2):
            return np.mean(coef, axis=0)
        return coef

    def explain_local(self, evaluation_examples, **kwargs):
        """Use LinearExplainer to get the local feature importances from the trained explainable model.

        :param evaluation_examples: The evaluation examples to compute local feature importances for.
        :type evaluation_examples: numpy.array or pandas.DataFrame or scipy.sparse.csr_matrix
        :return: The local explanation of feature importances.
        :rtype: Union[list, numpy.ndarray]
        """
        if self._sgd_explainer is None:
            self._sgd_explainer = _create_linear_explainer(self._sgd, self.multiclass, self._background,
                                                           self.covariance, self.random_state)
        return _compute_local_shap_values(self._sgd_explainer, evaluation_examples, self._classification)

    @property
    def expected_values(self):
        """Use LinearExplainer to get the expected values.

        :return: The expected values of the linear model.
        :rtype: list
        """
        if self._sgd_explainer is None:
            self._sgd_explainer = _create_linear_explainer(self._sgd, self.multiclass, self._background,
                                                           self.covariance, self.random_state)
        if isinstance(self._sgd_explainer, list):
            expected_values = []
            for explainer in self._sgd_explainer:
                expected_values.append(explainer.expected_value)
            return expected_values
        else:
            expected_values = self._sgd_explainer.expected_value
            if self._classification and not self.multiclass:
                expected_values = [-expected_values, expected_values]
            return expected_values

    @property
    def model(self):
        """Retrieve the underlying model.

        :return: The SGD model, either classifier or regressor.
        :rtype: Union[SGDClassifier, SGDRegressor]
        """
        return self._sgd
